import unittest
from collections.abc import Hashable, Iterator
from fontParts.base import FontPartsError


class TestSegment(unittest.TestCase):
    def getSegment_line(self):
        contour, unrequested = self.objectGenerator("contour")
        unrequested.append(contour)
        contour.appendPoint((0, 0), "move")
        contour.appendPoint((101, 202), "line")
        segment = contour[1]
        return segment

    def getSegment_offcurves(self):
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((0, 0), "offcurve")
        contour.appendPoint((100, 0), "offcurve")
        contour.appendPoint((100, 100), "offcurve")
        contour.appendPoint((0, 100), "offcurve")
        segment = contour[0]
        return segment

    # ----
    # Repr
    # ----

    def test_reprContents(self):
        segment = self.getSegment_line()
        result = segment._reprContents()
        self.assertIn(segment.type, result)
        self.assertIn(f"index={segment.index}", result)

    def test_reprContents_no_index(self):
        segment, _ = self.objectGenerator("segment")
        result = segment._reprContents()
        self.assertIn(segment.type, result)

    # -------
    # Parents
    # -------

    def getGlyph_contour(self):
        glyph, _ = self.objectGenerator("glyph")
        pen = glyph.getPen()
        pen.moveTo((100, -10))
        pen.lineTo((100, 100))
        pen.lineTo((200, 100))
        pen.lineTo((200, 0))
        pen.closePath()
        return glyph

    def test_get_contour(self):
        contour, _ = self.objectGenerator("contour")
        segment, _ = self.objectGenerator("segment")
        segment._contour = lambda: contour
        self.assertEqual(segment.contour, contour)

    def test_get_contour_orphan_segment(self):
        segment, _ = self.objectGenerator("segment")
        self.assertIsNone(segment.contour)

    def test_set_conotur(self):
        contour, _ = self.objectGenerator("contour")
        segment, _ = self.objectGenerator("segment")
        segment.contour = contour
        self.assertEqual(segment._contour(), contour)

    def test_set_contour_value_set(self):
        contour, _ = self.objectGenerator("contour")
        segment, _ = self.objectGenerator("segment")
        segment._contour = contour
        with self.assertRaises(AssertionError):
            segment.contour = contour

    def test_set_contour_value_none(self):
        segment, _ = self.objectGenerator("segment")
        segment.contour = None
        self.assertIsNone(segment._contour)

    def test_get_glyph(self):
        glyph = self.getGlyph_contour()
        segment = glyph[0][0]
        self.assertEqual(segment.glyph, glyph)

    def test_get_glyph_orphan_segment(self):
        segment, _ = self.objectGenerator("segment")
        self.assertIsNone(segment.glyph)

    def test_get_layer(self):
        layer, _ = self.objectGenerator("layer")
        layer["A"] = self.getGlyph_contour()
        segment = layer["A"][0][0]
        self.assertEqual(segment.layer, layer)

    def test_get_layer_orphan_segment(self):
        segment, _ = self.objectGenerator("segment")
        self.assertIsNone(segment.layer)

    def test_get_font(self):
        font, _ = self.objectGenerator("font")
        font["A"] = self.getGlyph_contour()
        segment = font["A"][0][0]
        self.assertEqual(segment.font, font)

    def test_get_font_orphan_segment(self):
        segment, _ = self.objectGenerator("segment")
        self.assertIsNone(segment.font)

    # -----
    # Index
    # -----

    def test_get_index(self):
        segment = self.getSegment_line()
        self.assertEqual(segment.index, 1)

    def test_get_index_orphan_segment(self):
        segment, _ = self.objectGenerator("segment")
        self.assertIsNone(segment.index)

    # ----
    # Type
    # ----

    def test_type_get(self):
        segment = self.getSegment_line()
        self.assertEqual(segment.type, "line")

    def test_set_move(self):
        segment = self.getSegment_line()
        segment.type = "move"
        self.assertEqual(segment.type, "move")

    def test_len_move(self):
        segment = self.getSegment_line()
        segment.type = "move"
        self.assertEqual(len(segment.points), 1)

    def test_oncuve_type_move(self):
        segment = self.getSegment_line()
        segment.type = "move"
        self.assertEqual(segment.onCurve.type, "move")

    def test_oncuve_x_y(self):
        segment = self.getSegment_line()
        segment.type = "move"
        self.assertEqual((segment.onCurve.x, segment.onCurve.y), (101, 202))

    def test_set_curve(self):
        segment = self.getSegment_line()
        segment.type = "curve"
        self.assertEqual(segment.type, "curve")

    def test_len_curve(self):
        segment = self.getSegment_line()
        segment.type = "curve"
        self.assertEqual(len(segment.points), 3)

    def test_curve_pt_types(self):
        segment = self.getSegment_line()
        segment.type = "curve"
        types = tuple(point.type for point in segment.points)
        self.assertEqual(types, ("offcurve", "offcurve", "curve"))

    def test_curve_pt_x_y(self):
        segment = self.getSegment_line()
        segment.type = "curve"
        coordinates = tuple((point.x, point.y) for point in segment.points)
        self.assertEqual(coordinates, ((0, 0), (101, 202), (101, 202)))

    def test_set_qcurve(self):
        segment = self.getSegment_line()
        segment.type = "qcurve"
        self.assertEqual(segment.type, "qcurve")

    def test_len_qcurve(self):
        segment = self.getSegment_line()
        segment.type = "qcurve"
        self.assertEqual(len(segment.points), 3)

    def test_qcurve_pt_types(self):
        segment = self.getSegment_line()
        segment.type = "qcurve"
        types = tuple(point.type for point in segment.points)
        self.assertEqual(types, ("offcurve", "offcurve", "qcurve"))

    def test_qcurve_pt_x_y(self):
        segment = self.getSegment_line()
        segment.type = "qcurve"
        coordinates = tuple((point.x, point.y) for point in segment.points)
        self.assertEqual(coordinates, ((0, 0), (101, 202), (101, 202)))

    def test_set_invalid_segment_type_string(self):
        segment = self.getSegment_line()
        with self.assertRaises(ValueError):
            segment.type = "xxx"

    def test_set_invalid_segment_type_int(self):
        segment = self.getSegment_line()
        with self.assertRaises(TypeError):
            segment.type = 123

    def test_offCurve_only_segment(self):
        segment = self.getSegment_offcurves()
        self.assertEqual(len(segment.contour), 1)
        # onCurve is a dummy None value, telling this is an on-curve-less quad blob
        self.assertIsNone(segment.onCurve)
        self.assertEqual(segment.points, segment.offCurve)
        self.assertEqual(segment.type, "qcurve")

    def test_set_type_no_oncurve(self):
        segment = self.getSegment_offcurves()
        segment.type = "line"
        self.assertEqual(segment.type, "qcurve")

    def test_set_type_orphan_segment(self):
        segment, _ = self.objectGenerator("segment")
        with self.assertRaises(FontPartsError):
            segment.type = "move"

    def test_set_type_same_value(self):
        segment = self.getSegment_line()
        segment.type = "line"
        self.assertEqual(segment.type, "line")

    # ------
    # Smooth
    # ------

    def test_get_smooth(self):
        segment = self.getSegment_line()
        self.assertFalse(segment.smooth)

    def test_get_smooth_no_oncurve(self):
        segment = self.getSegment_offcurves()
        self.assertTrue(segment.smooth)

    def test_set_smooth(self):
        segment = self.getSegment_line()
        segment.smooth = False
        self.assertFalse(segment.smooth)

    def test_set_smooth_no_oncurve(self):
        segment = self.getSegment_offcurves()
        segment.smooth = False
        self.assertTrue(segment.smooth)

    # -------
    # OnCurve
    # -------

    def test_get_onCurve(self):
        segment = self.getSegment_line()
        result = segment.onCurve.position
        self.assertEqual(result, (101, 202))

    def test_get_onCurve_no_points(self):
        segment, _ = self.objectGenerator("segment")
        self.assertIsNone(segment.onCurve)

    # ----
    # Iter
    # ----

    def test_iter(self):
        segment = self.getSegment_line()
        points = [p.position for p in segment]
        self.assertIsInstance(iter(segment), Iterator)
        self.assertEqual(points, [(101, 202)])

    # ---
    # Len
    # ---

    def test_len(self):
        segment = self.getSegment_line()
        self.assertEqual(len(segment), 1)

    def test_len_orphan_segment(self):
        segment, _ = self.objectGenerator("segment")
        self.assertEqual(len(segment), 0)

    # ----
    # Hash
    # ----

    def test_hash(self):
        segment = self.getSegment_line()
        self.assertEqual(isinstance(segment, Hashable), False)

    # --------
    # Equality
    # --------

    def test_object_equal_self(self):
        segment_one = self.getSegment_line()
        self.assertEqual(segment_one, segment_one)

    def test_object_not_equal_other(self):
        segment_one = self.getSegment_line()
        segment_two = self.getSegment_line()
        self.assertNotEqual(segment_one, segment_two)

    def test_object_equal_self_variable_assignment(self):
        segment_one = self.getSegment_line()
        a = segment_one
        self.assertEqual(segment_one, a)

    def test_object_not_equal_other_variable_assignment(self):
        segment_one = self.getSegment_line()
        segment_two = self.getSegment_line()
        a = segment_one
        self.assertNotEqual(segment_two, a)

    def test_equal_invalid_type(self):
        segment_one = self.getSegment_line()
        segment_two = "invalidType"
        result = segment_one.__eq__(segment_two)
        self.assertEqual(result, NotImplemented)

    # ---------
    # Selection
    # ---------

    def test_selected_true(self):
        segment = self.getSegment_line()
        try:
            segment.selected = False
        except NotImplementedError:
            return
        segment.selected = True
        self.assertEqual(segment.selected, True)

    def test_selected_false(self):
        segment = self.getSegment_line()
        try:
            segment.selected = False
        except NotImplementedError:
            return
        self.assertEqual(segment.selected, False)

    # -------------
    # Interpolation
    # -------------

    def test_isCompatible(self):
        segment1 = self.getSegment_line()
        segment2 = self.getSegment_line()
        segment2.type = "curve"
        compatible, report = segment1.isCompatible(segment2)
        self.assertTrue(compatible)
        self.assertFalse(report.fatal)
        self.assertFalse(report.typeDifference)

    def test_isCompatible_inconvertible_types(self):
        segment1 = self.getSegment_line()
        segment2 = self.getSegment_offcurves()
        compatible, report = segment1.isCompatible(segment2)
        self.assertFalse(compatible)
        self.assertTrue(report.fatal)
        self.assertTrue(report.typeDifference)

    # -----
    # Round
    # -----

    def test_round(self):
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((0, 10.4), "move")
        contour.appendPoint((10.5, 10.6), "line")
        segment = contour[1]
        segment.round()
        points = segment[0]
        self.assertEqual(points.position, (11, 11))



from __future__ import annotations
from typing import TYPE_CHECKING, Any

import defcon
from fontParts.base import BaseContour
from fontParts.base.annotations import BoundingBox, CoordinateLike
from fontParts.fontshell.base import RBaseObject
from fontParts.fontshell.point import RPoint
from fontParts.fontshell.segment import RSegment
from fontParts.fontshell.bPoint import RBPoint

if TYPE_CHECKING:
    from fontParts.base import BasePoint


class RContour(RBaseObject, BaseContour):
    wrapClass = defcon.Contour
    pointClass = RPoint
    segmentClass = RSegment
    bPointClass = RBPoint

    # --------------
    # Identification
    # --------------

    # index

    def _set_index(self, value: int) -> None:
        contour = self.naked()
        glyph = contour.glyph
        if glyph is not None:
            glyph.removeContour(contour)
            glyph.insertContour(value, contour)

    # identifier

    def _get_identifier(self) -> str | None:
        return self.naked().identifier

    def _getIdentifier(self) -> str:
        return self.naked().generateIdentifier()

    def _getIdentifierForPoint(self, point: BasePoint) -> str:
        contour = self.naked()
        nakedPoint = point.naked()
        return contour.generateIdentifierForPoint(nakedPoint)

    # ----
    # Open
    # ----

    def _get_open(self) -> bool:
        return self.naked().open

    # ------
    # Bounds
    # ------

    def _get_bounds(self) -> BoundingBox | None:
        return self.naked().bounds

    # ----
    # Area
    # ----

    def _get_area(self) -> float | None:
        value = self.naked().area
        if not value:
            return None
        return value

    # ---------
    # Direction
    # ---------

    def _get_clockwise(self) -> bool:
        return self.naked().clockwise

    def _reverse(self, **kwargs: Any) -> None:
        self.naked().reverse()

    # ------------------------
    # Point and Contour Inside
    # ------------------------

    def _pointInside(self, point: CoordinateLike) -> bool:
        return self.naked().pointInside(point)

    def _contourInside(self, otherContour: BaseContour) -> bool:
        return self.naked().contourInside(otherContour.naked(), segmentLength=5)

    # ------
    # Points
    # ------

    def _lenPoints(self, **kwargs: Any) -> int:
        return len(self.naked())

    def _getPoint(self, index: int, **kwargs: Any) -> RPoint:
        contour = self.naked()
        point = contour[index]
        return self.pointClass(point)

    def _insertPoint(
        self,
        index: int,
        position: CoordinateLike,
        type: str | None = None,
        smooth: bool | None = None,
        name: str | None = None,
        identifier: str | None = None,
        **kwargs: Any,
    ) -> None:
        point = self.pointClass()
        point.x = position[0]
        point.y = position[1]
        point.type = type
        point.smooth = smooth
        point.name = name
        nakedPoint = point.naked()
        if nakedPoint is not None:
            point = nakedPoint
        point.identifier = identifier
        contour = self.naked()
        contour.insertPoint(index, point)

    def _removePoint(self, index: int, preserveCurve: bool, **kwargs: Any) -> None:
        contour = self.naked()
        point = contour[index]
        contour.removePoint(point)



from __future__ import annotations
from typing import Any
from collections.abc import ItemsView

import defcon
from fontParts.base.annotations import KerningPairLike
from fontParts.base import BaseKerning
from fontParts.fontshell.base import RBaseObject


class RKerning(RBaseObject, BaseKerning):
    wrapClass = defcon.Kerning

    def _items(self) -> ItemsView[str, int]:
        return self.naked().items()

    def _contains(self, key: str) -> bool:
        return key in self.naked()

    def _setItem(self, key: str, value: int) -> None:
        self.naked()[key] = value

    def _getItem(self, key: str) -> Any:
        return self.naked()[key]

    def _delItem(self, key: str) -> None:
        del self.naked()[key]

    def _find(self, pair: KerningPairLike, default: int | float | None = 0) -> int:
        return self.naked().find(pair, default)



from __future__ import annotations
from typing import TYPE_CHECKING, Any
from collections.abc import Callable, Generator

from fontParts.base.errors import FontPartsError
from fontParts.base.base import (
    BaseObject,
    TransformationMixin,
    InterpolationMixin,
    SelectionMixin,
    dynamicProperty,
    reference,
)
from fontParts.base import normalizers
from fontParts.base.deprecated import DeprecatedSegment, RemovedSegment
from fontParts.base.compatibility import SegmentCompatibilityReporter
from fontParts.base.annotations import AffineTransformationLike, CollectionType

if TYPE_CHECKING:
    from fontParts.base.point import BasePoint
    from fontParts.base.contour import BaseContour
    from fontParts.base.glyph import BaseGlyph
    from fontParts.base.layer import BaseLayer
    from fontParts.base.font import BaseFont


class BaseSegment(
    BaseObject,
    TransformationMixin,
    InterpolationMixin,
    SelectionMixin,
    DeprecatedSegment,
    RemovedSegment,
):
    """Represent the basis for a segment object."""

    def _setPoints(self, points: CollectionType[BasePoint]) -> None:
        if hasattr(self, "_points"):
            raise AssertionError("segment has points")
        self._points = points

    def _reprContents(self) -> list[str]:
        contents = [f"{self.type}"]
        if self.index is not None:
            contents.append(f"index={self.index}")
        return contents

    # this class should not be used in hashable
    # collections since it is dynamically generated.

    __hash__ = None  # type: ignore[assignment]

    # -------
    # Parents
    # -------

    # Contour

    _contour: Callable[[], BaseContour] | None = None

    contour: dynamicProperty = dynamicProperty(
        "contour",
        """Get or set the segment's parent contour object.

        The value must be a :class:`BaseContour` instance or :obj:`None`.

        :return: The :class:`BaseContour` instance containing the segment
            or :obj:`None`.
        :raises AssertionError: If attempting to set the contour when it
            has already been set.

        Example::

            >>> contour = segment.contour

        """,
    )

    def _get_contour(self) -> BaseContour | None:
        if self._contour is None:
            return None
        return self._contour()

    def _set_contour(
        self, contour: BaseContour | Callable[[], BaseContour] | None
    ) -> None:
        if self._contour is not None:
            raise AssertionError("contour for segment already set")
        if contour is not None:
            contour = reference(contour)
        self._contour = contour

    # Glyph

    glyph: dynamicProperty = dynamicProperty(
        "glyph",
        """Get the segment's parent glyph object.

        This property is read-only.

        The value must be a :class:`BaseGlyph` instance or :obj:`None`.

        :return: The :class:`BaseGlyph` instance containing the segment
            or :obj:`None`.

        Example::

            >>> glyph = segment.glyph

        """,
    )

    def _get_glyph(self) -> BaseGlyph | None:
        if self._contour is None:
            return None
        return self.contour.glyph

    # Layer

    layer: dynamicProperty = dynamicProperty(
        "layer",
        """Get the segment's parent layer object.

        This property is read-only.

        :return: The :class:`BaseLayer` instance containing the segemnt
            or :obj:`None`.

        Example::

            >>> layer = segment.layer

        """,
    )

    def _get_layer(self) -> BaseLayer | None:
        if self._contour is None:
            return None
        return self.glyph.layer

    # Font

    font: dynamicProperty = dynamicProperty(
        "font",
        """Get the segment's parent font object.

        This property is read-only.

        :return: The :class:`BaseFont` instance containing the segment
            or :obj:`None`.

        Example::

            >>> font = segment.font

        """,
    )

    def _get_font(self) -> BaseFont | None:
        if self._contour is None:
            return None
        return self.glyph.font

    # --------
    # Equality
    # --------

    def __eq__(self, other: object) -> bool:
        """Check for equality with another segment.

        The :meth:`BaseObject.__eq__` method can't be used here
        because the :class:`BaseContour` implementation contructs
        segment objects without assigning an underlying `naked`
        object. Therefore, comparisons will always fail. This
        method overrides the base method and compares the
        :class:`BasePoint` contained by the segment.

        :param other: The segment to compare with.
        :return: :obj:`True` if the segments are equal, :obj:`False` otherwise.

        .. note::

            Subclasses may override this method.

        """
        if isinstance(other, self.__class__):
            return self.points == other.points
        return NotImplemented

    # --------------
    # Identification
    # --------------

    index: dynamicProperty = dynamicProperty(
        "base_index",
        """Get the index of the segment.

        This property is read-only.

        :return: An :class:`int` representing the segment's index within an
            ordered list of the parent contour's segments, or :obj:`None` if the
            segment does not belong to a contour.

        Example::

            >>> segment.index
            0

        """,
    )

    def _get_base_index(self) -> int | None:
        if self.contour is None:
            return None
        value = self._get_index()
        normalizedValue = normalizers.normalizeIndex(value)
        return normalizedValue

    def _get_index(self) -> int:
        """Get the index of the native segment.

        This is the environment implementation of the :attr:`BaseSegment.index`
        property getter.

        :return: An :class:`int` representing the segment's index within an
            ordered list of the parent contour's segments, or :obj:`None` if the
            segment does not belong to a contour.

        .. note::

            Subclasses may override this method.

        """
        contour = self.contour
        value = contour.segments.index(self)
        return value

    # ----------
    # Attributes
    # ----------

    type: dynamicProperty = dynamicProperty(
        "base_type",
        """Get or set the segment's type.

        The value must be a :class:`str` containing one of the following
        alternatives:

        +----------------+---------------------------------+
        | Type           | Description                     |
        +----------------+---------------------------------+
        | ``'move'``     | An on-curve move to.            |
        | ``'line'``     | An on-curve line to.            |
        | ``'curve'``    | An on-curve cubic curve to.     |
        | ``'qcurve'``   | An on-curve quadratic curve to. |
        +----------------+---------------------------------+

        :return: A :class:`str` representing the type of the segment.

        """,
    )

    def _get_base_type(self) -> str:
        value = self._get_type()
        value = normalizers.normalizeSegmentType(value)
        return value

    def _set_base_type(self, value: str) -> None:
        value = normalizers.normalizeSegmentType(value)
        self._set_type(value)

    def _get_type(self) -> str:
        """Get the native segment's type.

        This is the environment implementation of the :attr:`BaseSegment.type`
        property getter.

        :return: A :class:`str` representing the type of the segment. The value
         will have been normalized with :func:`normalizers.normalizeSegmentType`.

        .. note::

            Subclasses may override this method.

        """
        onCurve = self.onCurve
        if onCurve is None:
            return "qcurve"
        return onCurve.type

    def _set_type(self, newType: str) -> None:
        """Set the native segment's type.

        This is the environment implementation of the :attr:`BaseSegment.type`
        property setter.

        :param newType: The segment type definition as a :class:`str`. The value
            will have been normalized with :func:`normalizers.normalizeSegmentType`.
        :raises FontPartsError: If the segment does not belong to a contour.

        .. note::

            Subclasses may override this method.

        """
        contour = self.contour
        if contour is None:
            raise FontPartsError("The segment does not belong to a contour.")
        oldType = self.type
        if oldType == newType:
            return
        if self.onCurve is None:
            # special case with a single qcurve segment
            # and only offcurves, don't convert
            return
        # converting line <-> move
        if newType in ("move", "line") and oldType in ("move", "line"):
            pass
        # converting to a move or line
        elif newType not in ("curve", "qcurve"):
            offCurves = self.offCurve
            for point in offCurves:
                contour.removePoint(point)
        # converting a line/move to a curve/qcurve
        else:
            segments = contour.segments
            i = segments.index(self)
            prev = segments[i - 1].onCurve
            on = self.onCurve
            x = on.x
            y = on.y
            points = contour.points
            i = points.index(on)
            contour.insertPoint(i, (x, y), "offcurve")
            off2 = contour.points[i]
            contour.insertPoint(i, (prev.x, prev.y), "offcurve")
            off1 = contour.points[i]
            del self._points
            self._setPoints((off1, off2, on))
        self.onCurve.type = newType

    smooth: dynamicProperty = dynamicProperty(
        "base_smooth",
        """Get or set the segment's smooth state.

        The value must be a :class:`bool` indicating the segment's smooth state.

        :return: :obj:`True` if the segment is smooth, :obj:`False` if it is sharp.

        Example::

            >>> segment.smooth
            False
            >>> segment.smooth = True

        """,
    )

    def _get_base_smooth(self) -> bool:
        value = self._get_smooth()
        value = normalizers.normalizeBoolean(value)
        return value

    def _set_base_smooth(self, value: bool) -> None:
        value = normalizers.normalizeBoolean(value)
        self._set_smooth(value)

    def _get_smooth(self) -> bool:
        """Get the native segment's smooth state.

        This is the environment implementation of the :attr:`BaseSegment.smooth`
        property getter.

        :return: :obj:`True` if the segment is smooth, :obj:`False` if it is
            sharp. The value will have been normalized
            with :func:`normalizers.normalizeBoolean`.

        .. note::

            Subclasses may override this method.

        """
        onCurve = self.onCurve
        if onCurve is None:
            return True
        return onCurve.smooth

    def _set_smooth(self, value: bool) -> None:
        """Set the native segment's smooth state.

        This is the environment implementation of the :attr:`BaseSegment.smooth`
        property setter.

        :param value: The point's smooth state as a :class:`bool`. The value
            will have been normalized with :func:`normalizers.normalizeBoolean`.

        .. note::

            Subclasses may override this method.

        """
        onCurve = self.onCurve
        if onCurve is not None:
            self.onCurve.smooth = value

    # ------
    # Points
    # ------

    def __getitem__(self, index: int) -> BasePoint:
        """Get the point at the specified index.

        :param index: The zero-based index of the point to retrieve as
            an :class:`int`.
        :return: The :class:`BasePoint` instance located at the specified `index`.
        :raises IndexError: If the specified `index` is out of range.

        """
        return self._getItem(index)

    def _getItem(self, index: int) -> BasePoint:
        """Get the native point at the specified index.

        This is the environment implementation of :meth:`BaseSegment.__getitem__`.

        :param index: The zero-based index of the point to retrieve as
            an :class:`int`.
        :return: The :class:`BasePoint` instance located at the specified `index`.
        :raises IndexError: If the specified `index` is out of range.

        .. note::

            Subclasses may override this method.

        """
        return self.points[index]

    def __iter__(self):
        """Return an iterator over the points in the segment.

        :return: An iterator over the :class:`BasePoint` instances belonging to
            the segment.

        """
        return self._iterPoints()

    def _iterPoints(self, **kwargs: Any) -> Generator[BasePoint]:
        """Return an iterator over the points in the native segment.

        This is the environment implementation of :meth:`BaseSegment.__iter__`.

        :return: An iterator over the :class:`BasePoint` subclass instances
            belonging to the segment.

        .. note::

            Subclasses may override this method.

        """
        points = self.points
        count = len(points)
        index = 0
        while count:
            yield points[index]
            count -= 1
            index += 1

    def __len__(self) -> int:
        """Return the number of points in the segment.

        :return: An :class:`int` representing the number of :class:`BasePoint`
            instances belonging to the segment.

        """
        return self._len()

    def _len(self, **kwargs: Any) -> int:
        """Return the number of points in the native segment.

        This is the environment implementation of :meth:`BaseSegment.__len__`.

        :return: An :class:`int` representing the number of :class:`BasePoint`
            subclass instances belonging to the segment.

        .. note::

            Subclasses may override this method.

        """
        return len(self.points)

    points: dynamicProperty = dynamicProperty(
        "base_points",
        """Get a list of all points in the segment.

        This attribute is read-only.

        :return: A :class:`tuple` of :class`BasePoint` instances.

        """,
    )

    def _get_base_points(self) -> tuple[BasePoint, ...]:
        return self._get_points()

    def _get_points(self) -> tuple[BasePoint, ...]:
        """Get a list of all points in the native segment.

        This is the environment implementation of the :attr:`BaseSegment.points`
        property getter.

        :return: A :class:`tuple` of :class`BasePoints`.

        .. note::

            Subclasses may override this method.

        """
        if not hasattr(self, "_points"):
            return ()
        return tuple(self._points)

    onCurve: dynamicProperty = dynamicProperty(
        "base_onCurve",
        """Get the on-curve point in the segment.

        This property is read-only.

        :return: An on-curve :class:`BasePoint` instance or :obj:`None`.

        """,
    )

    def _get_base_onCurve(self) -> BasePoint | None:
        return self._get_onCurve()

    def _get_onCurve(self) -> BasePoint | None:
        """Get the on-curve point in the native segment.

        This is the environment implementation of
        the :attr:`BaseSegment.onCurve` property getter.

        :return: An on-curve :class:`BasePoint` instance or :obj:`None`.

        .. note::

            Subclasses may override this method.

        """
        points = self.points
        if not points:
            return None
        value = points[-1]
        if value.type == "offcurve":
            return None
        return value

    offCurve: dynamicProperty = dynamicProperty(
        "base_offCurve",
        """Get the off-curve points in the segment.

        This property is read-only.

        :return: An off-curve :class:`BasePoint` instance or :obj:`None`.

        """,
    )

    def _get_base_offCurve(self) -> tuple[BasePoint, ...]:
        return self._get_offCurve()

    def _get_offCurve(self) -> tuple[BasePoint, ...]:
        """Get the off-curve points in the native segment.

        This is the environment implementation of
        the :attr:`BaseSegment.offCurve` property getter.

        :return: An off-curve :class:`BasePoint` instance or :obj:`None`.

        .. note::

            Subclasses may override this method.

        """
        if self.points and self.points[-1].type == "offcurve":
            return self.points
        return self.points[:-1]

    # --------------
    # Transformation
    # --------------

    def _transformBy(self, matrix: AffineTransformationLike, **kwargs: Any):
        r"""Transform the native object according to the given matrix.

        This is the environment implementation of :meth:`TransformationMixin.transformBy`.

        :param matrix: The :ref:`type-transformation` to apply. The value will have
            been normalized with :func:`normalizers.normalizeTransformationMatrix`.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        for point in self.points:
            point.transformBy(matrix)

    # -------------
    # Interpolation
    # -------------

    compatibilityReporterClass = SegmentCompatibilityReporter

    def isCompatible(
        self, other: BaseSegment, cls=None
    ) -> tuple[bool, SegmentCompatibilityReporter]:
        """Evaluate interpolation compatibility with another segment.

        This method will return a :class:`bool` indicating if the segment is
        compatible for interpolation with `other`, and a :class:`str`
        containing compatibility notes.

        :param other: The other :class:`BaseSegment` instance to check
            compatibility with.
        :return: A :class:`tuple` where the first element is a :class:`bool`
            indicating compatibility, and the second element is
            a :class:`fontParts.base.compatibility.SegmentCompatibilityReporter`
            instance.

        Example::

            >>> compatible, report = self.isCompatible(otherSegment)
            >>> compatible
            False
            >>> compatible
            [Fatal] Segment: [0] + [0]
            [Fatal] Segment: [0] is line | [0] is move
            [Fatal] Segment: [1] + [1]
            [Fatal] Segment: [1] is line | [1] is qcurve

        """
        return super().isCompatible(other, BaseSegment)

    def _isCompatible(
        self, other: BaseSegment, reporter: SegmentCompatibilityReporter
    ) -> None:
        """Evaluate interpolation compatibility with another native segment.

        This is the environment implementation of :meth:`BaseSegment.isCompatible`.

        :param other: The other :class:`BaseSegment` instance to check
            compatibility with.
        :param reporter: An object used to report compatibility issues.

        .. note::

            Subclasses may override this method.

        """
        segment1 = self
        segment2 = other
        # type
        if segment1.type != segment2.type:
            # line <-> curve can be converted
            if {segment1.type, segment2.type} != {"curve", "line"}:
                reporter.typeDifference = True
                reporter.fatal = True

    # ----
    # Misc
    # ----

    def round(self) -> None:
        """Round coordinates in all the segment's points."""
        for point in self.points:
            point.round()



from __future__ import annotations

from typing import TYPE_CHECKING
from collections import Counter
from fontTools.misc.fixedTools import otRound
from pathlib import Path
import datetime
import os

from fontParts.base.annotations import (
    InterpolationFactorLike,
    InterpolationFactor,
    SkewAngleLike,
    SkewAngle,
    ScaleFactorLike,
    ScaleFactorPair,
    ScaleFactor,
    KerningPairLike,
    KerningPair,
    AffineTransformationLike,
    AffineTransformation,
    BoundingBox,
    BoundingBoxLike,
    Coordinate,
    CoordinateLike,
    RGBA,
    RGBALike,
    T,
    CollectionType,
    IntFloatType,
    LibValueType,
)

if TYPE_CHECKING:
    from fontParts.base.layer import BaseLayer
    from fontParts.base.glyph import BaseGlyph
    from fontParts.base.contour import BaseContour
    from fontParts.base.point import BasePoint
    from fontParts.base.segment import BaseSegment
    from fontParts.base.bPoint import BaseBPoint
    from fontParts.base.component import BaseComponent
    from fontParts.base.anchor import BaseAnchor
    from fontParts.base.guideline import BaseGuideline

# ----
# Font
# ----


def normalizeFileFormatVersion(value: int) -> int:
    """Normalize a font's file format version.

    :param value: The file format version of the font as an :class:`int`.
    :return: An :class:`int` representing the normalized file format version.
    :raises TypeError: If `value` is not of type :class:`int`.

    """
    if not isinstance(value, int):
        raise TypeError(
            f"Expected file format version 'value' to be of type int, not {type(value).__name__}."
        )
    return value


def normalizeFileStructure(value: str) -> str:
    """Normalize a font's file structure.

    :param value: The file structure to normalize as a :class:`str`.
    :return: A :class:`str` representing the normalized file structure.
    :raises TypeError: If `value` is not a :class:`str`.

    """
    allowedFileStructures = ["zip", "package"]
    if value not in allowedFileStructures:
        raise TypeError(
            f"File Structure must be {', '.join(allowedFileStructures)}, not {value}"
        )
    return value


def normalizeLayerOrder(value: CollectionType[str], font) -> tuple[str, ...]:
    """Normalize layer order.

    :param value: The layer order to normalize as a :class:`list`
        or :class:`tuple` of layer names. Each item in `value`:
        - must normalize with :func:`normalizeLayerName`.
        - must correspond to an existing layer name in the font.
        - must be unique (no duplicates are allowed).
    :param font: The :class:`BaseFont` (subclass) instance to which the
        normalization applies.
    :return: A :class:`tuple` of layer names in their normalized order.
    :raises TypeError:
        - If `value` is not a :class:`list` or :class:`tuple`.
        - If any `value` item is not a :class:`str`.
    :raises ValueError:
        - If `value` contains duplicate values.
        - If any `value` item does not exist in `font`.
        - If any `value` item is an empty :class:`str`.

    """
    if not isinstance(value, (tuple, list)):
        raise TypeError(f"Layer order must be a list, not {type(value).__name__}.")
    for v in value:
        normalizeLayerName(v)
    fontLayers = [layer.name for layer in font.layers]
    for name in value:
        if name not in fontLayers:
            raise ValueError(
                f"Layer must exist in font. {name} does not exist in font.layers."
            )
    duplicates = [v for v, count in Counter(value).items() if count > 1]
    if len(duplicates) != 0:
        raise ValueError(
            f"Duplicate layers are not allowed. Layer name(s) '{', '.join(duplicates)}' are duplicate(s)."
        )
    return tuple(value)


def normalizeDefaultLayerName(value: str, font) -> str:
    """Normalize a default layer name.

    :param value: The layer name to normalize as a :class:`str`. The value:
        - must normalize as layer name with :func:`normalizeLayerName`.
        - must be a layer in `font`.
    :param font: The :class:`BaseFont` (subclass) instance to which the
        normalization applies.
    :return: A :class:`str` representing the normalized default layer name.
    :raises TypeError: If `value` is not a :class:`str`.
    :raises ValueError:
        - If `value` does not exist in `font`.
        - If `value` is an empty :class:`str`.

    """
    value = normalizeLayerName(value)
    if value not in font.layerOrder:
        raise ValueError(f"No layer with the name '{value}' exists.")
    return str(value)


def normalizeGlyphOrder(value: CollectionType[str]) -> tuple[str, ...]:
    """Normalize glyph order.

    :param value: The glyph order to normalize as a :class:`list`
        or :class:`tuple` of glyph names. Each item in `value`:
        - must normalize as glyph names with :func:`normalizeGlyphName`.
        - must be unique (no duplicates are allowed).
    :return: A :class:`tuple` of glyph names in their normalized order.
    :raises TypeError:
        - If `value` is not a :class:`list` or :class:`tuple`.
        - If any `value` item is not a :class:`str`.
    :raises ValueError:
        - If `value` contains duplicate values.
        - If any `value` item is an empty :class:`str`.

    """
    if not isinstance(value, (tuple, list)):
        raise TypeError(f"Glyph order must be a list, not {type(value).__name__}.")
    for v in value:
        normalizeGlyphName(v)
    duplicates = sorted(v for v, count in Counter(value).items() if count > 1)
    if len(duplicates) != 0:
        raise ValueError(
            f"Duplicate glyph names are not allowed. Glyph name(s) '{', '.join(duplicates)}' are duplicate."
        )
    return tuple(value)


# -------
# Kerning
# -------


def normalizeKerningKey(value: KerningPairLike) -> KerningPair:
    """Normalize a kerning key.

    :param value: The kerning key to normalize as a :class:`tuple`
        or :class:`list` containing two items representing the left and right
        kerning key groups. Each item in `value`:
        - must be a :class:`str`.
        - must be at least one character long.
    :return: The normalized kerning key as a :class:`tuple` of :class:`str` items.
    :raises TypeError:
        - If `value` is not a :class:`list` or :class:`tuple`.
        - If any `value` item is not a :class:`str`.
    :raises ValueError:
        - If `value` does not a contain a pair of items.
        - If any `value` item is an empty :class:`str`.
        - If the left kerning key group starts with 'public.' but does not
          start with 'public.kern1.'.
        - If the right kerning key group starts with 'public.' but does not
          start with 'public.kern2.'.

    """
    if not isinstance(value, (tuple, list)):
        raise TypeError(
            f"Kerning key must be a tuple instance, not {type(value).__name__}."
        )
    if len(value) != 2:
        raise ValueError(
            f"Kerning key must be a tuple containing two items, not {len(value)}."
        )
    for v in value:
        if not isinstance(v, str):
            raise TypeError(
                f"Kerning key items must be strings, not {type(v).__name__}."
            )
        if len(v) < 1:
            raise ValueError("Kerning key items must be at least one character long")
    left, right = value
    if left.startswith("public.") and not left.startswith("public.kern1."):
        raise ValueError("Left Kerning key group must start with public.kern1.")
    if right.startswith("public.") and not right.startswith("public.kern2."):
        raise ValueError("Right Kerning key group must start with public.kern2.")
    return (left, right)


def normalizeKerningValue(value: IntFloatType) -> IntFloatType:
    """Normalize a kerning value.

    :param value: The kerning value to normalize as an :class:`int` or :class:`float`.
    :return: An :class:`int` or :class:`flaot` representing the normalized
        kerning value.
    raises TypeError: If `value` is not an :class:`int` or a :class:`float`.

    """
    if not isinstance(value, (int, float)):
        raise TypeError(
            f"Kerning value must be an int or a float, not {type(value).__name__}."
        )
    return value


# ------
# Groups
# ------


def normalizeGroupKey(value: str) -> str:
    """Normalize a group key.

    :param value: The group key to normalize as a non-empty :class:`str`.
    :return: A :class:`str` representing the normalized group key.
    :raises TyypeError: If `value` is not a :class:`str`.
    raises ValueError: If `value` is an empty :class:`str`.

    """
    if not isinstance(value, str):
        raise TypeError(f"Group key must be a string, not {type(value).__name__}.")
    if len(value) < 1:
        raise ValueError("Group key must be at least one character long.")
    return value


def normalizeGroupValue(value: CollectionType[str]) -> tuple[str, ...]:
    """Normalize a group value.

    :param value: The group value to normalize as a :class:`list`
        or :class:`tuple` of :class:`str` items representing the names of the
        glyphs in the group. All `value` items must
        normalize :func:`normalizeGlyphName`.
    :return: A :class:`tuple` of :class:`str` items representing the normalized
        gorup value.
    :raises TyypeError: If `value` is not a :class:`list` or :class:`tuple`.

    """
    if not isinstance(value, (tuple, list)):
        raise TypeError(f"Group value must be a list, not {type(value).__name__}.")
    value = [normalizeGlyphName(v) for v in value]
    return tuple(value)


# --------
# Features
# --------


def normalizeFeatureText(value: str) -> str:
    """Normalize feature text.

    :param value: The feature text to normalize as a :class:`str`.
    :return: A :class:`str` representing the normalized feature text.
    :raises TypeError: If `value` is not a :class:`str`.

    """
    if not isinstance(value, str):
        raise TypeError(f"Feature text must be a string, not {type(value).__name__}.")
    return value


# ---
# Lib
# ---


def normalizeLibKey(value: str) -> str:
    """Normalize a lib key.

    :param value: The lib key to normalize as a non-empty :class:`str`.
    :return: A :class:`str` representing the normalized lib key.
    :raises TypeError: If `value` is not a :class:`str`.
    :raises ValueError: If `value` is an empty :class:`str`.

    """
    if not isinstance(value, str):
        raise TypeError(f"Lib key must be a string, not {type(value).__name__}.")
    if len(value) < 1:
        raise ValueError("Lib key must be at least one character.")
    return value


def normalizeLibValue(value: LibValueType) -> LibValueType:
    """Normalize a lib value.

    If `value` is a collection (:class:`list`, :class:`tuple`, or :class:`dict`),
    its elements will be normalized recursively.

    :param value: The lib value to normalize. The value and any value items must
        be a type compatible with the :mod:`plistlib` module.
    :return: The normalized lib value, of the same type as `value`.
    :raises ValueError: If `value` or any of it's items is :obj:`None`.

    """
    acceptedTypes = (
        str,
        int,
        float,
        bool,
        tuple,
        list,
        dict,
        bytes,
        bytearray,
        datetime.datetime,
    )
    if not isinstance(value, acceptedTypes):
        acceptedTypesStr = (
            ", ".join(t.__name__ for t in acceptedTypes[:-1])
            + f" or {acceptedTypes[-1].__name__}"
        )
        raise TypeError(
            f"Lib value must be a plistlib-supported type ({acceptedTypesStr})"
            f", got {type(value).__name__}."
        )
    if isinstance(value, (list, tuple)):
        for v in value:
            normalizeLibValue(v)
    elif isinstance(value, dict):
        for k, v in value.items():
            normalizeLibKey(k)
            normalizeLibValue(v)
    return value


# -----
# Layer
# -----


def normalizeLayer(value: BaseLayer) -> BaseLayer:
    """Normalize a layer.

    :param value: The layer to normalize as an instance of :class:`BaseLayer`.
    :return: The normalized :class:`BaseLayer`instance.
    :raises TypeError: If `value` is not an instance of :class:`BaseLayer`.

    """
    from fontParts.base.layer import BaseLayer

    return normalizeInternalObjectType(value, BaseLayer, "Layer")


def normalizeLayerName(value: str) -> str:
    """Normalize a layer name.

    :param value: The layer name to normalize as a non-empty :class:`str`.
    :return: A :class:`str` representing the normalized layer name.
    :raises TypeError: If `value` is not a :class:`str`.
    :raises ValueError: If `value` is an empty :class:`str`.

    """
    if not isinstance(value, str):
        raise TypeError(f"Layer names must be strings, not {type(value).__name__}.")
    if len(value) < 1:
        raise ValueError("Layer names must be at least one character long.")
    return value


# -----
# Glyph
# -----


def normalizeGlyph(value: BaseGlyph) -> BaseGlyph:
    """Normalize a glyph.

    :param value: The glyph to normalize as an instance of :class:`BaseGlyph`.
    :return: The normalized :class:`BaseGlyph` instance.
    :raises TypeError: If `value` is not an instance of :class:`BaseGlyph`.

    """
    from fontParts.base.glyph import BaseGlyph

    return normalizeInternalObjectType(value, BaseGlyph, "Glyph")


def normalizeGlyphName(value: str) -> str:
    """Normalize a glyph's name.

    :param value: The glyph name to normalize as a non-empty :class:`str`.
    :return: A :class:`str` representing the normalized glyph name.
    :raises TypeError: If `value` is not a :class:`str`.
    :raises ValueError: If `value` is an empty string.

    """
    if not isinstance(value, str):
        raise TypeError(f"Glyph names must be strings, not {type(value).__name__}.")
    if len(value) < 1:
        raise ValueError("Glyph names must be at least one character long.")
    return value


def normalizeGlyphUnicodes(value: CollectionType[int]) -> tuple[int, ...]:
    """Normalize a glyph's unicodes.

    :param value: The glyph unicodes to normalize as a :class:`list`
        or :class:`tuple` of Unicode values. Each item in `value`:
        - must normalize as glyph unicodes with :func:`normalizeGlyphUnicode`.
        - must be unique (no duplicates are allowed).
    :return: A :class:`tuple` of :class:`int` values representing the normalized
        glyph unicodes.
    :raises TypeError: If `value` is not a :class:`list` or :class:`tuple`.
    :raises ValueError:
        - If `value` contains duplicate values.
        - If any `value` item is not a valid hexadecimal value.
        - If any `value` item is not within the unicode range.

    """
    if not isinstance(value, (tuple, list)):
        raise TypeError(f"Glyph unicodes must be a list, not {type(value).__name__}.")
    values = [normalizeGlyphUnicode(v) for v in value]
    duplicates = [v for v, count in Counter(value).items() if count > 1]
    if len(duplicates) != 0:
        raise ValueError("Duplicate unicode values are not allowed.")
    return tuple(values)


def normalizeGlyphUnicode(value: int | str) -> int:
    """Normalize a glyph's unicode.

    :param value: The glyph Unicode value to normalize as an :class:`int` or a
        hexadecimal :class:`str`. The value must be within the unicode range.
    :return: An :class:`int` representing the normalized unicode value.
    :raises TypeError: If `value` is not and :class:`int` or a :class:`str`.
    :raises ValueError:
        - If `value` is not a valid hexadecimal value.
        - If `value` is not within the unicode range.

    """
    if not isinstance(value, (int, str)) or isinstance(value, bool):
        raise TypeError(
            f"Glyph unicode must be an int or hex string, not {type(value).__name__}."
        )
    if isinstance(value, str):
        try:
            value = int(value, 16)
        except ValueError:
            raise ValueError("Glyph unicode hex must be a valid hex string.")
    if value < 0 or value > 1114111:
        raise ValueError("Glyph unicode must be in the Unicode range.")
    return value


def normalizeGlyphWidth(value: IntFloatType) -> IntFloatType:
    """Normalize a glyph's width.

    :param value: The glyph width to normalize as an :class:`int` or :class:`float`.
    :return: The normalized glyph width, of the same type as `value`.
    :raises TypeError: if `value` is not an :class:`int` or a :class:`float`.

    """
    if not isinstance(value, (int, float)):
        raise TypeError(
            f"Glyph width must be int or float, not {type(value).__name__}."
        )
    return value


def normalizeGlyphLeftMargin(value: IntFloatType | None) -> IntFloatType | None:
    """Normalize a glyph's left margin.

    :param value: The glyph left margin to normalize as
        an :class:`int`, :class:`float`, or :obj:`None`.
    :return: The normalized glyph left margin, of the same type as `value`.
    :raises TypeError: if `value` is not an :class:`int` or a :class:`float`.

    """
    if not isinstance(value, (int, float)) and value is not None:
        raise TypeError(
            f"Glyph left margin must be int or float, not {type(value).__name__}."
        )
    return value


def normalizeGlyphRightMargin(value: IntFloatType | None) -> IntFloatType | None:
    """Normalize a glyph's right margin.

    :param value: The glyph right margin to normalize as
        an :class:`int`, :class:`float`, or :obj:`None`.
    :return: The normalized glyph right margin, of the same type as `value`.
    :raises TypeError: if `value` is not an :class:`int` or a :class:`float`.

    """
    if not isinstance(value, (int, float)) and value is not None:
        raise TypeError(
            f"Glyph right margin must be int or float, not {type(value).__name__}."
        )
    return value


def normalizeGlyphHeight(value: IntFloatType) -> IntFloatType:
    """Normalize a glyph's height.

    :param value: The glyph height to normalize as an :class:`int` or :class:`float`.
    :return: The normalized glyph height, of the same type as `value`.
    :raises TypeError: if `value` is not an :class:`int` or a :class:`float`.

    """
    if not isinstance(value, (int, float)):
        raise TypeError(
            f"Glyph height must be int or float, not {type(value).__name__}."
        )
    return value


def normalizeGlyphBottomMargin(value: IntFloatType | None) -> IntFloatType | None:
    """Normalize a glyph's bottom margin.

    :param value: The glyph bottom margin to normalize as
        an :class:`int`, :class:`float`, or :obj:`None`.
    :return: The normalized glyph bottom margin, of the same type as `value`.
    :raises TypeError: if `value` is not an :class:`int` or a :class:`float`.

    """
    if not isinstance(value, (int, float)) and value is not None:
        raise TypeError(
            f"Glyph bottom margin must be int or float, not {type(value).__name__}."
        )
    return value


def normalizeGlyphTopMargin(value: IntFloatType | None) -> IntFloatType | None:
    """Normalize a glyph's top margin.

    :param value: The glyph top margin to normalize as
        an :class:`int`, :class:`float`, or :obj:`None`.
    :return: The normalized glyph top margin, of the same type as `value`.
    :raises TypeError: If `value` is not an :class:`int` or a :class:`float`.

    """
    if not isinstance(value, (int, float)) and value is not None:
        raise TypeError(
            f"Glyph top margin must be int or float, not {type(value).__name__}."
        )
    return value


def normalizeGlyphFormatVersion(value: IntFloatType) -> int:
    """Normalize a glyph's format version for saving to XML string.

    :param value: The format version to normalize as
        an :class:`int` or a :class:`float` equal to either ``1`` or ``2``.
    :return: An :class:`int` representing the normalized glyph format version.
    :raises TypeError: If `value` is not an :class:`int` or a :class:`float`.
    :raises ValueError: If `value` is not equal to ``1`` or ``2``.

    """
    if not isinstance(value, (int, float)):
        raise TypeError(
            f"Glyph Format Version must be int or float, not {type(value).__name__}."
        )
    value = int(value)
    if value not in (1, 2):
        raise ValueError(f"Glyph Format Version must be either 1 or 2, not {value}.")
    return value


# -------
# Contour
# -------


def normalizeContour(value: BaseContour) -> BaseContour:
    """Normalize a contour.

    :param value: The contour to normalize as an instance of :class:`BaseContour`.
    :return: The normalized :class:`BaseContour` instance.
    :raises TypeError: If `value` is not an instance of :class:`BaseContour`.

    """
    from fontParts.base.contour import BaseContour

    return normalizeInternalObjectType(value, BaseContour, "Contour")


# -----
# Point
# -----


def normalizePointType(value: str) -> str:
    """Normalize a point type.

    :param value: The point type to normalize as a :class:`str` containing one
        of the following values:

        +----------------+---------------------------------------------------+
        | Value          | Description                                       |
        +----------------+---------------------------------------------------+
        | ``'move'``     | The first point in an open contour.               |
        | ``'line'``     | A straight line from the previous point.          |
        | ``'offcurve'`` | A control point in a curve or qcurve.             |
        | ``'curve'``    | A cubic Bézier curve from the previous point.     |
        | ``'qcurve'``   | A quadratic Bézier curve from the previous point. |
        +----------------+---------------------------------------------------+

    :return: A :class:`str` representing the normalized point type.
    :raises TypeError: If `value` is not a :class:`str`.
    :raises ValueError: If `value` is not one of the allowed types.

    """
    allowedTypes = ["move", "line", "offcurve", "curve", "qcurve"]
    if not isinstance(value, str):
        raise TypeError(f"Point type must be a string, not {type(value).__name__}.")
    if value not in allowedTypes:
        raise ValueError(
            "Point type must be '{}'; not {!r}.".format(
                "', '".join(allowedTypes), value
            )
        )
    return value


def normalizePointName(value: str) -> str:
    """Normalize a point name.

    :param value: The point name to normalize as a non-empty :class:`str`.
    :return: A :class:`str` representing the normalized point name.
    :raises TypeError: If `value` is not a :class:`str`.
    :raises ValueError: If `value` is an empty :class:`str`.

    """
    if not isinstance(value, str):
        raise TypeError(f"Point names must be strings, not {type(value).__name__}.")
    if len(value) < 1:
        raise ValueError("Point names must be at least one character long.")
    return value


def normalizePoint(value: BasePoint) -> BasePoint:
    """Normalize a point.

    :param value: The contour to normalize as an instance of :class:`BasePoint`.
    :return: The normalized :class:`BasePoint` instance.
    :raises TypeError: If `value` is not an instance of :class:`BasePoint`.

    """
    from fontParts.base.point import BasePoint

    return normalizeInternalObjectType(value, BasePoint, "Point")


# -------
# Segment
# -------


def normalizeSegment(value: BaseSegment) -> BaseSegment:
    """Normalize a segment.

    :param value: The contour to normalize as an instance of :class:`BaseSegment`.
    :return: The normalized :class:`BaseSegment` instance.
    :raises TypeError: If `value` is not an instance of :class:`BaseSegment`.

    """
    from fontParts.base.segment import BaseSegment

    return normalizeInternalObjectType(value, BaseSegment, "Segment")


def normalizeSegmentType(value: str) -> str:
    """Normalize a segment type.

    :param value: The segment type to normalize as a :class:`str` containing one
        of the following values:

        +--------------+-------------------------------------------------+
        | Value        | Description                                     |
        +--------------+-------------------------------------------------+
        | ``'move'``   | The start of an open contour.                   |
        | ``'line'``   | A straight segment between two on-curve points. |
        | ``'curve'``  | A cubic Bézier curve segment.                   |
        | ``'qcurve'`` | A quadratic Bézier curve segment.               |
        +--------------+-------------------------------------------------+

    :return: A :class:`str` representing the normalized segment type.
    :raises TypeError: If `value` is not a :class:`str`.
    :raises ValueError: If `value` is not one of the allowed types.

    """
    allowedTypes = ["move", "line", "curve", "qcurve"]
    if not isinstance(value, str):
        raise TypeError(f"Segment type must be a string, not {type(value).__name__}.")
    if value not in allowedTypes:
        raise ValueError(
            "Segment type must be '{}'; not {!r}.".format(
                "', '".join(allowedTypes), value
            )
        )
    return value


# ------
# BPoint
# ------


def normalizeBPoint(value: BaseBPoint) -> BaseBPoint:
    """Normalize a bPoint.

    :param value: The contour to normalize as an instance of :class:`BaseBPoint`.
    :return: The normalized :class:`BaseBPoint` instance.
    :raises TypeError: If `value` is not an instance of :class:`BaseBPoint`.

    """
    from fontParts.base.bPoint import BaseBPoint

    return normalizeInternalObjectType(value, BaseBPoint, "bPoint")


def normalizeBPointType(value: str) -> str:
    """Normalize a bPoint type.

    :param value: The bPoint type to normalize as a :class:`str` containing one
        of the following values:

        +--------------+---------------------------------------------------------+
        | Value        | Description                                             |
        +--------------+---------------------------------------------------------+
        | ``'curve'``  | A point where bcpIn and bcpOut are smooth (linked).     |
        | ``'corner'`` | A point where bcpIn and bcpOut are not smooth (linked). |
        +--------------+---------------------------------------------------------+

    :return: A :class:`str` representing the normalized bPoint type.
    :raises TypeError: If `value` is not a :class:`str`.
    :raises ValueError: If `value` is not one of the allowed types.

    """
    allowedTypes = ["corner", "curve"]
    if not isinstance(value, str):
        raise TypeError(f"bPoint type must be a string, not {type(value).__name__}.")
    if value not in allowedTypes:
        raise ValueError(f"bPoint type must be 'corner' or 'curve', not {value!r}.")
    return value


# ---------
# Component
# ---------


def normalizeComponent(value: BaseComponent) -> BaseComponent:
    """Normalize a component.

    :param value: The contour to normalize as an instance of :class:`BaseComponent`.
    :return: The normalized :class:`BaseComponent` instance.
    :raises TypeError: If `value` is not an instance of :class:`BaseComponent`.

    """
    from fontParts.base.component import BaseComponent

    return normalizeInternalObjectType(value, BaseComponent, "Component")


def normalizeComponentScale(value: ScaleFactorPair) -> ScaleFactor:
    """Normalize a component scale.

    :param value: The component scale to normalize as a :class:`list`
        or :class:`tuple` of two :class:`int` or :class:`float` values.
    :return: A :class:`tuple` of two :class:`float` values representing the
        normalized component scale.

    """
    if not isinstance(value, (list, tuple)):
        raise TypeError(
            f"Component scale must be a tuple instance, not {type(value).__name__}."
        )
    if not len(value) == 2:
        raise ValueError(
            f"Transformation scale tuple must contain two values, not {len(value)}."
        )
    for v in value:
        if not isinstance(v, (int, float)):
            raise TypeError(
                f"Transformation scale tuple values must be int or float, not {type(v).__name__}."
            )
    return (float(value[0]), float(value[1]))


# ------
# Anchor
# ------


def normalizeAnchor(value: BaseAnchor) -> BaseAnchor:
    """Normalize an anchor.

    :param value: The contour to normalize as an instance of :class:`BaseAnchor`.
    :return: The normalized :class:`BaseAnchor` instance.
    :raises TypeError: If `value` is not an instance of :class:`BaseAnchor`.

    """
    from fontParts.base.anchor import BaseAnchor

    return normalizeInternalObjectType(value, BaseAnchor, "Anchor")


def normalizeAnchorName(value: str) -> str:
    """Normalize an anchor name.

    :param value: The anchor name to normalize as a non-empty :class:`str`,
        or :obj:`None`.
    :return: A :class:`str` representing the normalized anchor name, or :obj:`None`.
    :raises TypeError: If `value` is not a :class:`str` or :obj:`None`.
    :raises ValueError: If `value` is an empty :class:`str`.

    """
    if value is None:
        return None
    if not isinstance(value, str):
        raise TypeError(f"Anchor names must be strings, not {type(value).__name__}.")
    if len(value) < 1:
        raise ValueError("Anchor names must be at least one character long or None.")
    return value


# ---------
# Guideline
# ---------


def normalizeGuideline(value: BaseGuideline) -> BaseGuideline:
    """Normalize a guideline.

    :param value: The contour to normalize as an instance of :class:`BaseGuideline`.
    :return: The normalized :class:`BaseGuideline` instance.
    :raises TypeError: If `value` is not an instance of :class:`BaseGuideline`.

    """
    from fontParts.base.guideline import BaseGuideline

    return normalizeInternalObjectType(value, BaseGuideline, "Guideline")


def normalizeGuidelineName(value: str) -> str:
    """Normalize a guideline name.

    :param value: The guideline name to normalize as a non-empty :class:`str`
    :return: A :class:`str` representing the normalized guideline name.
    :raises TypeError: If `value` is not a :class:`str`.
    :raises ValueError: If `value` is an empty :class:`str`.

    """
    if not isinstance(value, str):
        raise TypeError(f"Guideline names must be strings, not {type(value).__name__}.")
    if len(value) < 1:
        raise ValueError("Guideline names must be at least one character long.")
    return value


# -------
# Generic
# -------


def normalizeInternalObjectType(value: object, cls: type[T], name: str) -> T:
    """Normalize an internal object type.

    :param value: The object instance to normalize.
    :param cls: The class against which to check the type of `value`.
    :param name: The name of the variable being checked (for error messages).
    :return: The normalized object, of the same type as `value`.
    :raise TypeError: If `value` is not an instance of `cls`.

    """
    if not isinstance(value, cls):
        raise TypeError(
            f"{name} must be a {name} instance, not {type(value).__name__}."
        )
    return value


def normalizeBoolean(value: int) -> bool:
    """Normalize a boolean.

    :param value: The boolean to normalize as an :class:`int` of ``0`` or
        ``1``, or a :class:`bool`.
    :return: A :class:`bool` representing the normalized value.
    :raise ValueError: If `value` is not a valid :class:`bool` or :class:`int`.

    """
    if isinstance(value, int) and value in (0, 1):
        value = bool(value)
    if not isinstance(value, bool):
        raise ValueError(f"Boolean values must be True or False, not '{value}'.")
    return value


# Identification


def normalizeIndex(value: int | None) -> int | None:
    """Normalize an index.

    :param value: The index to normalize as an :class:`int`, or :obj:`None`.
    :return: The normalized index, of the same type as `value`.
    :raise ValueError: If `value` is not an :class:`int` or :obj:`None`.

    """
    if value is not None:
        if not isinstance(value, int):
            raise TypeError(
                f"Indexes must be None or integers, not {type(value).__name__}."
            )
    return value


def normalizeIdentifier(value: str | None) -> str | None:
    """Normalize an identifier.

    :param value: The identifier to normalize as a non-empty :class:`str`,
        or :obj:`None`. The value:
        - must not be longer than 100 characters.
        - must not contain a character out of the range ``0x20`` - ``0x7E``.
    :return: The normalized identifier, of the same type as `value`.
    :raises TypeError: If `value` is not a :class:`str` or :obj:`None`.
    :raises ValueError:
        - If `value` is an empty :class:`str`.
        - If `value` is longer than 100 characters.
        - If `value` contains a character outside the range ``0x20`` - ``0x7E``.

    """
    if value is None:
        return value
    if not isinstance(value, str):
        raise TypeError(f"Identifiers must be strings, not {type(value).__name__}.")
    if len(value) == 0:
        raise ValueError("The identifier string is empty.")
    if len(value) > 100:
        raise ValueError(
            f"The identifier string has a length ({len(value)}) greater than the maximum allowed (100)."
        )
    for c in value:
        v = ord(c)
        if v < 0x20 or v > 0x7E:
            raise ValueError(
                f"The identifier string ('{value}') contains a character outside of the range 0x20 - 0x7E."
            )
    return value


# Coordinates


def normalizeX(value: IntFloatType) -> IntFloatType:
    """Normalize an x-coordinate.

    :param value: The x-coordinate to normalize as an :class:`int` or :class:`float`.
    :return: The normalized glyph top margin, of the same type as `value`.
    :raises TypeError: If `value` is not an :class:`int` or a :class:`float`.

    """
    if not isinstance(value, (int, float)):
        raise TypeError(
            f"X-coordinates must be instances of int or float, not {type(value).__name__}."
        )
    return value


def normalizeY(value: IntFloatType) -> IntFloatType:
    """Normalize a y-coordinate.

    :param value: The y-coordinate to normalize as an :class:`int` or :class:`float`.
    :return: The normalized glyph top margin, of the same type as `value`.
    :raises TypeError: If `value` is not an :class:`int` or a :class:`float`.

    """
    if not isinstance(value, (int, float)):
        raise TypeError(
            f"Y-coordinates must be instances of int or float, not {type(value).__name__}."
        )
    return value


def normalizeCoordinateTuple(value: CoordinateLike) -> Coordinate:
    """Normalize a coordinate tuple.

    :param value: The coordinate tuple to normalize as a :class:`list`
        or :class:`tuple` of two :class:`int` or :class:`float` values.
    :return: :return: A :class:`tuple` of two values of the same type as the
         items in `value`, representing the normalized coordinates.
    :raises TypeError: If `value` is not a :class:`list` or :class:`tuple`.
    :raises ValueError: If `value` does not contain exactly two items.

    """
    if not isinstance(value, (tuple, list)):
        raise TypeError(
            f"Coordinates must be tuple instances, not {type(value).__name__}."
        )
    if len(value) != 2:
        raise ValueError(
            f"Coordinates must be tuples containing two items, not {len(value)}."
        )
    x, y = value
    x = normalizeX(x)
    y = normalizeY(y)
    return (x, y)


def normalizeBoundingBox(value: BoundingBoxLike) -> BoundingBox:
    """Normalize a bounding box.

    :param value: The bounding box to normalize as a :class:`list`
        or :class:`tuple` of four :class:`int` or :class:`float` values
        representing the coordinates (in order) xMin, yMin, xMax, and yMax. The
        xMin and yMin values must be less than or equal to the corresponding xMax,
        yMax values.
    :return: A :class:`tuple` of four :class:`float` values representing the
        normalized bounding box.
    :raises TypeError:
        - If `value` is not a :class:`tuple` or :class:`list`.
        - If any `value` item is not an :class:`int` or a :class:`float`.
    :raises ValueError:
        - If `value` does not contain exactly four items.
        - If xMin is greater than xMax.
        - If yMin is greater than yMax.

    """
    if not isinstance(value, (tuple, list)):
        raise TypeError(
            f"Bounding box must be tuple instances, not {type(value).__name__}."
        )
    if len(value) != 4:
        raise ValueError(
            f"Bounding box must be tuples containing four items, not {len(value)}."
        )
    for v in value:
        if not isinstance(v, (int, float)):
            raise TypeError(
                f"Bounding box values must be instances of int or float, not {type(v).__name__}."
            )
    xMin, yMin, xMax, yMax = value
    if xMin > xMax:
        raise ValueError("Bounding box xMin must be less than or equal to xMax.")
    if yMin > yMax:
        raise ValueError("Bounding box yMin must be less than or equal to yMax.")
    return (float(xMin), float(yMin), float(xMax), float(yMax))


def normalizeArea(value: IntFloatType) -> float:
    """Normalize an area.

    :param value: The area to normalize as a positive :class:`int` or :class:`float`.
    :return: A positive :class:`float` representing the normalized area.
    :raises TypeError: If `value` is not an :class:`int` or a :class:`float`.
    :raises ValueError: If `value` is not a positive value.

    """
    if not isinstance(value, (int, float)):
        raise TypeError(
            f"Area must be an instance of int or float, not {type(value).__name__}."
        )
    if value < 0:
        raise ValueError(f"Area must be a positive int or float, not {repr(value)}.")
    return float(value)


def normalizeRotationAngle(value: IntFloatType) -> float:
    """Normalize an angle.

    :param value: The angle to normalize as an :class:`int` or a :class:`float`
        between ``-360 and ``360``. If the value is negative, it is normalized by
        adding it to ``360``.
    :return: A :class:`float` between ``0.0`` and ``360.0`` representing the
        normalized rotation angle.
    :raises TypeError: If `value` is not an :class:`int` or a :class:`float`.
    :raises ValueError: If `value` is not between ``-360 and ``360``.

    """
    if not isinstance(value, (int, float)):
        raise TypeError(
            f"Angle must be instances of int or float, not {type(value).__name__}."
        )
    if abs(value) > 360:
        raise ValueError("Angle must be between -360 and 360.")
    if value < 0:
        value = value + 360
    return float(value)


# Color


def normalizeColor(value: RGBALike) -> RGBA:
    """Normalize a color.

    :param value: The color to normalize as a :class:`list` or :class:`tuple`
        of four :class:`int` or :class:`float` values between ``0`` and ``1``,
        corresponding to the red, green, blue, and alpha (RGBA) channels, in that
        order.
    :return: A :class:`tuple` of four :class:`float` values representing the
        normalized color.
    :raises TypeError:
        - If `value` is not a :class:`list` or :class:`tuple`.
        - If any `value` item is not an :class:`int` or a :class:`float`.
    :raises ValueError:
        - If `value` does not contain exactly four items.
        - If any `value` item is not between ``0`` and ``1``.

    """
    from fontParts.base.color import Color

    if not isinstance(value, (tuple, list, Color)):
        raise TypeError(f"Colors must be tuple instances, not {type(value).__name__}.")
    if not len(value) == 4:
        raise ValueError(f"Colors must contain four values, not {len(value)}.")
    for component, v in zip("rgba", value):
        if not isinstance(v, (int, float)):
            raise TypeError(
                f"The value for the {component} component ({v}) is not an int or float."
            )
        if v < 0 or v > 1:
            raise ValueError(
                f"The value for the {component} component ({v}) is not between 0 and 1."
            )
    r, g, b, a = value
    return (float(r), float(g), float(b), float(a))


# Note


def normalizeGlyphNote(value: str) -> str:
    """Normalize a glyph note.

    :param value: The glyph note to normalize as a :class:`str`.
    :return: A :class:`str` representing the normalized glyph note.
    :raises TypeError if `value` is not a :class:`str`.

    """
    if not isinstance(value, str):
        raise TypeError(f"Note must be a string, not {type(value).__name__}.")
    return value


# File Path


def normalizeFilePath(value: str | os.PathLike) -> str:
    """Normalize a file path.

    :param value: The file path to normalize as a :class:`str` or :class:`pathlib.Path`.
    :return: A :class:`str` representing the normalized file path.
    :raises TypeError if `value` is not a :class:`str` or :class:`pathlib.Path`.
    :raises FileNotFoundError: If the file path cannot be resolved because it does not exist.

    """
    if not isinstance(value, (str, Path)):
        raise TypeError(
            f"File paths must be strings or Path objects, not {type(value).__name__}."
        )
    return str(value)


# Interpolation


def normalizeInterpolationFactor(value: InterpolationFactorLike) -> InterpolationFactor:
    """Normalize an interpolation factor.

    :param value: The interpolation factor to normalize as a single :class:`int`
        or :class:`float`, or a :class:`tuple` or :class:`list` of
        two :class:`int` or :class:`float` values.
    :return: A :class:`tuple` of two :class:`float` values representing the
        normalized interpolation factor.
    :raises TypeError:
        - If `value` is not an :class:`int`, :class:`float`, :class:`tuple`,
          or :class:`list`.
        - If any `value` item is not an :class:`int` or a :class:`float`.
    :raises ValueError:
        - If `value` is a :class:`tuple` or :class:`list` and does not contain
          exactly two items.

    """
    if not isinstance(value, (int, float, list, tuple)):
        raise TypeError(
            f"Interpolation factor must be an int, float, tuple, or list, not {type(value).__name__}."
        )
    if isinstance(value, (int, float)):
        return (float(value), float(value))

    if len(value) != 2:
        raise ValueError(
            f"Interpolation factor tuple must contain two items, not {len(value)}."
        )
    for v in value:
        if not isinstance(v, (int, float)):
            raise TypeError(
                f"Interpolation factor tuple values must be instances of int or float, not {type(v).__name__}."
            )
    return (float(value[0]), float(value[1]))


# ---------------
# Transformations
# ---------------


def normalizeTransformationMatrix(
    value: AffineTransformationLike,
) -> AffineTransformation:
    """Normalize a transformation matrix.

    :param value: The transformation matrix to normalize as a :class:`list`
        or :class:`tuple` of six :class:`int` or :class:`float` values.
    :return: A :class:`tuple` of six :class:`float` values representing the
        normalized transformation matrix.
    :raises TypeError:
        - If `value` is not a :class:`tuple` or :class:`list`.
        - If any `value` item is not an :class:`int` or a :class:`float`.
    :raises ValueError: If `value` does not contain exactly six items.

    """
    if not isinstance(value, (tuple, list)):
        raise TypeError(
            f"Transformation matrices must be tuple instances, not {type(value).__name__}."
        )
    if not len(value) == 6:
        raise ValueError(
            f"Transformation matrices must contain six values, not {len(value)}."
        )
    for v in value:
        if not isinstance(v, (int, float)):
            raise TypeError(
                f"Transformation matrix values must be instances of int or float, not {type(v).__name__}."
            )
    a, b, c, d, e, f = value
    return (float(a), float(b), float(c), float(d), float(e), float(f))


def normalizeTransformationOffset(value: CoordinateLike) -> Coordinate:
    """Normalize a transformation offset.

    :param value: The transformation offset to normalize as a :class:`list`
        or :class:`tuple` of two :class:`int` or :class:`float` values.
    :return: A :class:`tuple` of two :class:`float` values representing the
        normalized transformation offset.
    :raises TypeError:
        - If `value` is not a :class:`tuple`.
        - If any `value` item is not an :class:`int` or a :class:`float`.
    :raises ValueError: If `value` does not contain exactly two items.

    """
    return normalizeCoordinateTuple(value)


def normalizeTransformationSkewAngle(value: SkewAngleLike) -> SkewAngle:
    """Normalize a transformation skew angle.

    :param value: The skew angle to normalize as a single :class:`int`
     or :class:`float`, or a :class:`tuple` or :class:`list` of
     two :class:`int` or :class:`float` values. Each value must be
     between ``-360 and ``360``. If the value is negative, it is normalized by
     adding it to ``360``.
    :return: A :class:`tuple` of two :class:`float` values between ``0.0`` and
        ``360.0`` representing the normalized skew angle.
    :raises TypeError:
        - If `value` is not an :class:`int`, :class:`float`, :class:`tuple`,
          or :class:`list`.
        - If any `value` item is not an :class:`int` or a :class:`float`.
    :raises ValueError:
        - If `value` is a :class:`tuple` or :class:`list` and does not contain
          exactly two items.
    :raises ValueError:
        - If `value` does not contain exactly two items.
        - If any `value` item is not between ``-360`` and ``360``.

    """
    if not isinstance(value, (int, float, list, tuple)):
        raise TypeError(
            f"Transformation skew angle must be an int, float, or tuple instances, not {type(value).__name__}."
        )
    if isinstance(value, (int, float)):
        value = (float(value), 0.0)
    else:
        if not len(value) == 2:
            raise ValueError(
                f"Transformation skew angle tuple must contain two values, not {len(value)}."
            )
    normalized = []
    for v in value:
        if not isinstance(v, (int, float)):
            raise TypeError(
                f"Transformation skew angle tuple values must be int or float, not {type(v).__name__}."
            )
        if not -360 <= v <= 360:
            raise ValueError("Transformation skew angle must be between -360 and 360.")
        normalized.append(float(v + 360) if v < 0 else float(v))
    return (normalized[0], normalized[1])


def normalizeTransformationScale(value: ScaleFactorLike) -> ScaleFactor:
    """Normalize a transformation scale.

    :param value: The scale to normalize as a single :class:`int`
        or :class:`float`, or a :class:`tuple` or :class:`list` of
        two :class:`int` or :class:`float` values.
    :return: A :class:`tuple` of two :class:`float` values representing the
        normalized scale.
    :raises TypeError:
        - If `value` is not an :class:`int`, :class:`float`, :class:`tuple`,
          or :class:`list`.
        - If any `value` item is not an :class:`int` or a :class:`float`.
    :raises ValueError:
        - If `value` is a :class:`tuple` or :class:`list` and does not contain
          exactly two items.

    """
    if not isinstance(value, (int, float, list, tuple)):
        raise TypeError(
            f"Transformation scale must be an int, float, or tuple instances, not {type(value).__name__}."
        )
    if isinstance(value, (int, float)):
        return (float(value), float(value))

    if not len(value) == 2:
        raise ValueError(
            f"Transformation scale tuple must contain two values, not {len(value)}."
        )
    for v in value:
        if not isinstance(v, (int, float)):
            raise TypeError(
                f"Transformation scale tuple values must be int or float, not {type(v).__name__}."
            )
    return (float(value[0]), float(value[1]))


def normalizeVisualRounding(value: IntFloatType) -> int:
    """Normalize rounding.

    Python 3 uses banker's rounding, meaning anything that is at 0.5 will round
    to the nearest even number. This isn't always ideal for point coordinates,
    so instead, this function rounds to the higher number
    with :func:`fontTools.misc.roundTools.otRound`.

    :param value: The value to round as an :class:`int` or a :class:`float`.
    :return: An :class:`int` representing the normalized rounding.
    :raises TypeError: If `value` is not an :class:`int` or a :class:`float`.

    """
    if not isinstance(value, (int, float)):
        raise TypeError(
            f"Value to round must be an int or float, not {type(value).__name__}."
        )
    return otRound(value)



from __future__ import annotations
from typing import TYPE_CHECKING
from collections.abc import Callable, Iterator
from collections.abc import MutableMapping

from fontParts.base.base import BaseDict, dynamicProperty, reference
from fontParts.base import normalizers
from fontParts.base.deprecated import DeprecatedLib, RemovedLib
from fontParts.base.annotations import LibValueType

if TYPE_CHECKING:
    from fontParts.base.glyph import BaseGlyph
    from fontParts.base.font import BaseFont
    from fontParts.base.layer import BaseLayer
    from fontParts.base.base import BaseItems
    from fontParts.base.base import BaseValues
    from fontParts.base.base import BaseKeys


class BaseLib(BaseDict, DeprecatedLib, RemovedLib):
    """Represent the basis for a lib object.

    This object behaves like a Python :class:`dict` object. Most of the
    dictionary functionality comes from :class:`BaseDict`. Consult that
    object's documentation for the required environment implementation
    details.

    :cvar KeyNormalizer: A function to normalize the key of the dictionary.
        Defaults to :func:`normalizers.normalizeLibKey`.
    :cvar ValueNormalizer: A function to normalize the value of the dictionary.
        Defaults to :func:`normalizers.normalizeLibValue`.

    This object is normally created as part of a :class:`BaseFont`.
    An orphan :class:`BaseLib` object instance can be created like this::

        >>> lib = RLib()

    """

    keyNormalizer: Callable[[str], str] = normalizers.normalizeLibKey
    valueNormalizer: Callable[[LibValueType], LibValueType] = (
        normalizers.normalizeLibValue
    )

    def _reprContents(self) -> list[str]:
        contents = []
        if self.glyph is not None:
            contents.append("in glyph")
            contents += self.glyph._reprContents()
        if self.font:
            contents.append("in font")
            contents += self.font._reprContents()
        return contents

    # -------
    # Parents
    # -------

    # Glyph

    _glyph: Callable[[], BaseGlyph] | None = None

    glyph: dynamicProperty = dynamicProperty(
        "glyph",
        """Get or set the lib's parent glyph object.

        The value must be a :class:`BaseGlyph` instance or :obj:`None`.

        :return: The :class:`BaseGlyph` instance containing the lib
            or :obj:`None`.
        :raises AssertionError:
            - If the font for the lib has already been set.
            - If attempting to set the glyph when it has already been set and is
              not the same as the provided glyph.

        Example::

            >>> glyph = lib.glyph

        """,
    )

    def _get_glyph(self) -> BaseGlyph | None:
        if self._glyph is None:
            return None
        return self._glyph()

    def _set_glyph(self, glyph: BaseGlyph | Callable[[], BaseGlyph] | None) -> None:
        if self._font is not None:
            raise AssertionError("font for lib already set")
        if self._glyph is not None and self._glyph() != glyph:
            raise AssertionError("glyph for lib already set and is not same as glyph")
        if glyph is not None:
            glyph = reference(glyph)
        self._glyph = glyph

    # Font

    _font: Callable[[], BaseFont] | None = None

    font: dynamicProperty = dynamicProperty(
        "font",
        """Get or set the lib's parent font object.

        The value must be a :class:`BaseFont` instance or :obj:`None`.

        :return: The :class:`BaseFont` instance containing the lib
            or :obj:`None`.
        :raises AssertionError:
            - If attempting to set the font when it has already been set and is
              not the same as the provided font.
            - If the glyph for the lib has already been set.

        Example::

            >>> font = lib.font

        """,
    )

    def _get_font(self) -> BaseFont | None:
        if self._font is not None:
            return self._font()
        elif self._glyph is not None:
            return self.glyph.font
        return None

    def _set_font(self, font: BaseFont | Callable[[], BaseFont] | None) -> None:
        if self._font is not None and self._font() != font:
            raise AssertionError("font for lib already set and is not same as font")
        if self._glyph is not None:
            raise AssertionError("glyph for lib already set")
        if font is not None:
            font = reference(font)
        self._font = font

    # Layer

    layer: dynamicProperty = dynamicProperty(
        "layer",
        """Get the lib's parent layer object.

        This property is read-only.

        :return: The :class:`BaseLayer` instance containing the contour
            or :obj:`None`.

        Example::

            >>> layer = lib.layer

        """,
    )

    def _get_layer(self) -> BaseLayer | None:
        if self._glyph is None:
            return None
        return self.glyph.layer

    # ---------------------
    # RoboFab Compatibility
    # ---------------------

    def remove(self, key: str) -> None:
        """Remove the specified key from the lib.

        :param key: The key to remove as a :class:`str`.

        .. note::

            This is a backwards compatibility method.

        Example::

            >>> font.lib.remove('myKey')

        """
        del self[key]

    def asDict(self) -> dict[str, LibValueType]:
        """Return the lib as a dictionary.

        :return A :class:`dict` reflecting the contents of the lib.

        .. note::

            This is a backwards compatibility method.

        Example::

            >>> font.lib.asDict()

        """
        return dict(self)

    # -------------------
    # Inherited Functions
    # -------------------

    def __contains__(self, key: str) -> bool:
        """Check if the given key exists in the lib.

        :param key: The key to check for existence as a :class:`str`.
        :return: :obj:`True` if the `key` exists in the lib :obj:`False` otherwise.

        Example::

            >>> "public.glyphOrder" in font.lib
            True

        """
        return super().__contains__(key)

    def __delitem__(self, key: str) -> None:
        """Remove the given key from the lib.

        :param key: The key to remove as a :class:`str`.

        Example::

            >>> del font.lib["public.glyphOrder"]

        """
        super().__delitem__(key)

    def __getitem__(self, key: str) -> LibValueType:
        """Get the value associated with the given key.

        :param key: The key to retrieve the value for as a :class:`str`.
        :return: The :ref:`type-lib-value` associated with the specified key.
        :raise KeyError: If the specified `key` does not exist.

        Example::

            >>> font.lib["public.glyphOrder"]
            ["A", "B", "C"]

        .. note::

            Any changes to the returned lib contents will not be reflected in
            it's :class:`BaseLib` instance. To make changes to this content,
            do the following::

                >>> lib = font.lib["public.glyphOrder"]
                >>> lib.remove("A")
                >>> font.lib["public.glyphOrder"] = lib

        """
        return super().__getitem__(key)

    def __iter__(self) -> Iterator[str]:
        """Return an iterator over the keys in the lib.

        The iteration order is not fixed.

        :return: An :class:`Iterator` over the :class:`str` keys.

        Example::

            >>> for key in font.lib:
            >>>     print key
            "public.glyphOrder"
            "org.robofab.scripts.SomeData"
            "public.postscriptNames"

        """
        return super().__iter__()

    def __len__(self) -> int:
        """Return the number of keys in the lib.

        :return: An :class:`int` representing the number of keys in the lib.

        Example::

            >>> len(font.lib)
            5

        """
        return super().__len__()

    def __setitem__(self, key: str, value: LibValueType) -> None:
        """Set the value for a given key in the lib.

        :param key: The key to set as a :class:`str`.
        :param value: The :ref:`type-lib-value` to set for the given key.

        Example::

            >>> font.lib["public.glyphOrder"] = ["A", "B", "C"]

        """
        super().__setitem__(key, value)

    def clear(self) -> None:
        """Remove all keys from the lib.

        This will reset the :class:`BaseLib` instance to an empty dictionary.

        Example::

            >>> font.lib.clear()

        """
        super().clear()

    def get(self, key: str, default: LibValueType | None = None) -> LibValueType | None:
        """Get the value for the given key in the lib.

        If the given `key` is not found, The specified `default` will be returned.

        :param key: The key to look up as a :class:`str`.
        :param default: The optional default :ref:`type-lib-value` to return if
            the `key` is not found. Defaults to :obj:`None`.
        :return: The :ref:`type-lib-value` for the given `key`, or the `default`
            value if the `key` is not found.

        Example::

            >>> font.lib.get("public.glyphOrder")
            ["A", "B", "C"]
            >>> font.lib.get("missingKey", default="Default Value")
            "Default Value"

        ..note::

            Any changes to the returned lib contents will not be reflected in
            it's :class:`BaseLib` instance. To make changes to this content,
            do the following::

                >>> lib = font.lib.get("public.glyphOrder")
                >>> lib.remove("A")
                >>> font.lib["public.glyphOrder"] = lib

        """
        return super().get(key, default)

    def items(self) -> BaseItems[str, LibValueType]:
        """Return the lib's items.

        Each item is represented as a :class:`tuple` of key-value pairs, where:
            - `key` is a :class:`str`.
            - `value` is a :ref:`type-lib-value`.

        :return: A :ref:`type-view` of the lib's ``(key, value)`` pairs.

        Example::

            >>> font.lib.items()
            [("public.glyphOrder", ["A", "B", "C"]),
             ("public.postscriptNames", {'be': 'uni0431', 'ze': 'uni0437'})]

        """
        return super().items()

    def keys(self) -> BaseKeys[str]:
        """Return the lib's keys.

        :return: A :ref:`type-view` of :class:`str` items representing the lib's keys.

        Example::

            >>> font.lib.keys()
            ["public.glyphOrder", "org.robofab.scripts.SomeData",
             "public.postscriptNames"]

        """
        return super().keys()

    def values(self) -> BaseValues[LibValueType]:
        """Return the lib's values.

        :return: A :ref:`type-view` of :ref:`type-lib-value <lib values>`.

        Example::

            >>> font.lib.items()
            [["A", "B", "C"], {'be': 'uni0431', 'ze': 'uni0437'}]

        """
        return super().values()

    def pop(self, key: str, default: LibValueType | None = None) -> LibValueType | None:
        """Remove the specified key and return its associated value.

        If the `key` does not exist, the `default` value is returned.

        :param key: The key to remove as a :class:`str`.
        :param default: The optional default :ref:`type-lib-value` to return if
            the `key` is not found. Defaults to :obj:`None`.
        :return: The :ref:`type-lib-value` associated with the given `key`, or
            the `default` value if the `key` is not found.

        Example::

            >>> font.lib.pop("public.glyphOrder")
            ["A", "B", "C"]

        """
        return super().pop(key, default)

    def update(self, otherLib: MutableMapping[str, LibValueType]) -> None:
        """Update the current lib with key-value pairs from another.

        For each key in `otherLib`:
            - If the key exists in the current lib, its value is replaced with
              the value from `otherLib`.
            - If the key does not exist in the current lib, it is added.

        Keys that exist in the current lib but are not in `otherLib` remain unchanged.

        :param otherLib: A :class:`MutableMapping` of :class:`str` keys mapped
            to :ref:`type-lib-value <lib values>` to update the current lib with.

        Example::

            >>> font.lib.update(newLib)

        """
        super().update(otherLib)



from __future__ import annotations
from typing import TYPE_CHECKING
from collections.abc import Callable, Iterator
from collections.abc import MutableMapping

from fontParts.base.base import BaseDict, dynamicProperty, reference
from fontParts.base import normalizers
from fontParts.base.deprecated import DeprecatedGroups, RemovedGroups
from fontParts.base.annotations import CollectionType

if TYPE_CHECKING:
    from fontParts.base.font import BaseFont
    from fontParts.base.base import BaseKeys
    from fontParts.base.base import BaseItems
    from fontParts.base.base import BaseValues

ValueType = tuple[str, ...]
GroupsDict = dict[str, ValueType]


class BaseGroups(BaseDict, DeprecatedGroups, RemovedGroups):
    """Represent the basis for a groups object.

    This object behaves like a Python :class:`dict` object. Most of the
    dictionary functionality comes from :class:`BaseDict`. Consult that
    object's documentation for the required environment implementation
    details.

    :cvar keyNormalizer: A function to normalize the key of the dictionary.
        Defaults to :func:`normalizers.normalizeGroupKey`
    :cvar valueNormalizer: A function to normalize the value of the dictionary.
        Defaults to :func:`normalizers.normalizeGroupValue`

    This object is normally created as part of a :class:`BaseFont`.
    An orphan :class:`BaseGroups` object instance can be created like this::

        >>> groups = RGroups()

    """

    keyNormalizer: Callable[[str], str] = normalizers.normalizeGroupKey
    valueNormalizer: Callable[[CollectionType[str]], ValueType] = (
        normalizers.normalizeGroupValue
    )

    def _reprContents(self) -> list[str]:
        contents = []
        if self.font is not None:
            contents.append("for font")
            contents += self.font._reprContents()
        return contents

    # -------
    # Parents
    # -------

    # Font

    _font: Callable[[], BaseFont] | None = None

    font: dynamicProperty = dynamicProperty(
        "font",
        """Get or set the groups' parent font object.

        The value must be a :class:`BaseFont` instance or :obj:`None`.

        :return: The :class:`BaseFont` instance containing the group
            or :obj:`None`.
        :raises AssertionError:
            - If attempting to set the font when it has already been set and is
              not the same as the provided font.

        Example::

            >>> font = groups.font

        """,
    )

    def _get_font(self) -> BaseFont | None:
        if self._font is None:
            return None
        return self._font()

    def _set_font(self, font: BaseFont | Callable[[], BaseFont] | None) -> None:
        if self._font is not None and self._font != font:
            raise AssertionError("font for groups already set and is not same as font")
        if font is not None:
            font = reference(font)
        self._font = font

    # ---------
    # Searching
    # ---------

    def findGlyph(self, glyphName: str) -> list[str]:
        """Retrieve the groups associated with the given glyph.

        :param glyphName: The name of the glyph to search for as a :class:`str`.
        :return: A :class:`list` of :class:`str` items.


        Example::

            >>> font.groups.findGlyph("A")
            ["A_accented"]

        """
        glyphName = normalizers.normalizeGlyphName(glyphName)
        groupNames = self._findGlyph(glyphName)
        return [self._normalizeKey(groupName) for groupName in groupNames]

    def _findGlyph(self, glyphName: str) -> list[str]:
        """Retrieve the groups associated with the given native glyph.

        This is the environment implementation of :meth:`BaseGroups.findGlyph`.

        :param glyphName: The name of the glyph to search for as a :class:`str`.
            The value will have been normalized
            with :func:`normalizers.normalizeGlyphName`.
        :return: A :class:`list` of :class:`str` items.

        .. note::

            Subclasses may override this method.

        """
        found = []
        for key, groupList in self.items():
            if glyphName in groupList:
                found.append(key)
        return found

    # --------------
    # Kerning Groups
    # --------------

    side1KerningGroups: dynamicProperty = dynamicProperty(
        "base_side1KerningGroups",
        """Get all groups marked as potential side 1 (left) kerning members.

        This property is read-only.

        :return: A :class:`dict` of :class:`str` group names mapped
            to a :class:`tuple` of :class:`str` glyph names.

        Example::

            >>> side1Groups = groups.side1KerningGroups

        """,
    )

    def _get_base_side1KerningGroups(self) -> GroupsDict:
        kerningGroups = self._get_side1KerningGroups()
        normalized = {}
        for name, members in kerningGroups.items():
            name = normalizers.normalizeGroupKey(name)
            members = normalizers.normalizeGroupValue(members)
            normalized[name] = members
        return normalized

    def _get_side1KerningGroups(self) -> GroupsDict:
        """Get all native groups marked as potential side 1 (left) kerning members.

        This is the environment implementation of the
        :attr:`BaseGroups.side1KerningGroups` property getter.

        :return: A :class:`dict` of :class:`str` group names mapped
            to a :class:`tuple` of :class:`str` glyph names.

        .. note::

            Subclasses may override this method.

        """
        found = {}
        for name, contents in self.items():
            if name.startswith("public.kern1."):
                found[name] = contents
        return found

    side2KerningGroups: dynamicProperty = dynamicProperty(
        "base_side2KerningGroups",
        """Get all groups marked as potential side 2 (right) kerning members.

        This property is read-only.

        :return: A :class:`dict` of :class:`str` group names mapped
            to a :class:`tuple` of :class:`str` glyph names.

        Example::

            >>> side2Groups = groups.side2KerningGroups

        """,
    )

    def _get_base_side2KerningGroups(self) -> GroupsDict:
        kerningGroups = self._get_side2KerningGroups()
        normalized = {}
        for name, members in kerningGroups.items():
            name = normalizers.normalizeGroupKey(name)
            members = normalizers.normalizeGroupValue(members)
            normalized[name] = members
        return normalized

    def _get_side2KerningGroups(self) -> GroupsDict:
        """Get all native groups marked as potential side 2 (right) kerning members.

        This is the environment implementation of the
        :attr:`BaseGroups.side2KerningGroups` property getter.

        :return: A :class:`dict` of :class:`str` group names mapped
            to a :class:`tuple` of :class:`str` glyph names.

        .. note::

            Subclasses may override this method.

        """
        found = {}
        for name, contents in self.items():
            if name.startswith("public.kern2."):
                found[name] = contents
        return found

    # ---------------------
    # RoboFab Compatibility
    # ---------------------

    def remove(self, groupName: str) -> None:
        """Remove the given group from the current groups.

        :param: groupName: The name of the group to be removed as a :class:`str`.

        .. note::

            This is a backwards compatibility method.

        Example::

            >>> font.groups.remove("myKey")

        """
        del self[groupName]

    def asDict(self) -> GroupsDict:
        """Return the groups as a dictionary.

        :return A :class:`dict` reflecting the contents of the current groups.

        .. note::

            This is a backwards compatibility method.

        Example::

            >>> font.groups.asDict()

        """
        return dict(self)

    # -------------------
    # Inherited Functions
    # -------------------

    def __contains__(self, groupName: str) -> bool:
        """Check if the given key exists in the groups.

        :param groupName: The group name to check for existence as a :class:`str`.
        :return: :obj:`True` if the `groupName` exists in the
            groups, :obj:`False` otherwise.

        Example::

            >>> "myGroup" in font.groups
            True

        """
        return super().__contains__(groupName)

    def __delitem__(self, groupName: str) -> None:
        """Remove the given group from the current groups instance.

        :param groupName: The name of the group to remove as a :class:`str`.

        Example::

            >>> del font.groups["myGroup"]

        """
        super().__delitem__(groupName)

    def __getitem__(self, groupName: str) -> tuple[str, ...]:
        """Get the contents of the given group.

        :param groupName: The group name to retrieve the value for as a :class:`str`.
        :return: A :class:`tuple` of :class:`str` glyph names.
        :raise KeyError: If the specified `groupName` does not exist.

        Example::

            >>> font.groups["myGroup"]
            ("A", "B", "C")

        .. note::

            Any changes to the returned lib contents will not be reflected in
            it's :class:`BaseGroups` instance. To make changes to this content,
            do the following::

                >>> group = font.groups["myGroup"]
                >>> group.remove("A")
                >>> font.groups["myGroup"] = group

        """
        return super().__getitem__(groupName)

    def __iter__(self) -> Iterator[str]:
        """Return an iterator over the keys in the current groups instance.

        The iteration order is not fixed.

        :return: An :class:`Iterator` over the :class:`str` keys.

        Example::

            >>> for groupName in font.groups:
            >>>     print groupName
            "myGroup"
            "myGroup3"
            "myGroup2"

        """
        return super().__iter__()

    def __len__(self) -> int:
        """Return the number of groups in the current groups instance.

        :return: An :class:`int` representing the number of groups in the
            current groups instance.

        Example::

            >>> len(font.groups)
            5

        """
        return super().__len__()

    def __setitem__(self, groupName: str, glyphNames: CollectionType[str]) -> None:
        """Set the glyph names for a given group in the current groups instance.

        :param groupName: The group name to set as a :class:`str`.
        :param glyphNames: The glyph names to set for the given group as
            a :class:`list` or :class:`tuple` of :class:`str` items.

        Example::

            >>> font.groups["myGroup"] = ["A", "B", "C"]

        """

        super().__setitem__(groupName, glyphNames)

    def clear(self) -> None:
        """Remove all groups from the current groups instance.

        This will reset the :class:`BaseGroups` instance to an empty dictionary.

        Example::

            >>> font.groups.clear()

        """
        super().clear()

    def get(
        self, groupName: str, default: CollectionType[str] | None = None
    ) -> tuple[str, ...] | None:
        """Get the contents for the given group in the current groups instance.

        If the given `groupName` is not found, The specified `default` will be
        returned.

        :param groupName: The group name to look up as a :class:`str`.
        :param default: The optional default value to return if the `groupName`
            is not found`. The value must be either a class`list` or :class:`tuple`
            of :class:`str` glyph names, or :obj:`None`. Defaults to :obj:`None`.
        :return: The contents of the given group as a :class:`tuple`
            of :class:`str` items, or the `default` value if the group is not found.

        Example::

            >>> font.groups["myGroup"]
            ("A", "B", "C")

        ..note::

            Any changes to the returned lib contents will not be reflected in
            it's :class:`BaseGroups` instance. To make changes to this content,
            do the following::

                >>> group = font.groups["myGroup"]
                >>> group.remove("A")
                >>> font.groups["myGroup"] = group

        """
        return super().get(groupName, default)

    def items(self) -> BaseItems[str, ValueType]:
        """Return the items in the current groups instance.

        Each item is represented as a :class:`tuple` of key-value pairs, where:
            - `key` is a :class:`str` representing a group name.
            - `value` is a :class:`tuple` of :class:`str` glyph names.

        :return: A :ref:`type-view` of the groups' ``(key, value)`` pairs.

        Example::

            >>> font.groups.items()
            BaseGroups_items([("myGroup", ("A", "B", "C")),
            ("myGroup2", ("D", "E", "F")), ...])

        """
        return super().items()

    def keys(self) -> BaseKeys:
        """Return the group names (keys) in the current groups instance.

        :return: A :ref:`type-view` of :class:`str` items representing the groups' keys.

        Example::

            >>> font.groups.keys()
            BaseGroups_keys(["myGroup4", "myGroup1", "myGroup5", ...])

        """
        return super().keys()

    def values(self) -> BaseValues:
        """Return the values in the current groups instance.

        :return: A :ref:`type-view` of the groups' values as :class:`tuple`
            items of :class:`str` glyph names.

        Example::

            >>> font.groups.values()
            BaseGroups_values([("A", "B", "C"), ("D", "E", "F")]

        """
        return super().values()

    def pop(
        self, groupName: str, default: CollectionType[str] | None = None
    ) -> tuple[str, ...] | None:
        """Remove the specified group and return its associated contents.

        If the `groupName` does not exist, the `default` value is returned.

        :param groupName: The group to remove as a :class:`str`.
        :param default: The optional default value to return if the `groupName`
            is not found`. The value must be either a class`list` or :class:`tuple`
            of :class:`str` glyph names, or :obj:`None`. Defaults to :obj:`None`.
        :return: The contents of the given group as a :class:`tuple`
            of :class:`str` items, or the `default` value if the group is not found.

        Example::

            >>> font.groups.pop("myGroup")
            ("A", "B", "C")

        """
        return super().pop(groupName, default)

    def update(self, otherGroups: MutableMapping[str, CollectionType[str]]) -> None:
        """Update the current groups instance with key-value pairs from another.

        For each group in `otherGroups`:
            - If the group exists in the current groups instance, its value is
              replaced with the value from `otherGroups`.
            - If the key does not exist in the current groups instance, it is added.

        Keys that exist in the current groups instance but are not in `otherLib`
        remain unchanged.

        :param otherLib: A :class:`MutableMapping` of :class:`str` group names
            mapped to a :class:`tuple` of :class:`str` glyph names to update the
            current groups instance with.

        Example::

            >>> font.groups.update(newGroups)

        """
        super().update(otherGroups)



import unittest
import collections
from fontParts.base import FontPartsError


class TestAnchor(unittest.TestCase):
    def getAnchor_generic(self):
        anchor, _ = self.objectGenerator("anchor")
        anchor.name = "Anchor Attribute Test"
        anchor.x = 1
        anchor.y = 2
        anchor.color = None
        return anchor

    # ----
    # repr
    # ----

    def test_reprContents(self):
        anchor = self.getAnchor_generic()
        value = anchor._reprContents()
        self.assertIsInstance(value, list)
        for i in value:
            self.assertIsInstance(i, str)

    def test_reprContents_noGlyph(self):
        anchor, _ = self.objectGenerator("anchor")
        value = anchor._reprContents()
        self.assertIsInstance(value, list)
        for i in value:
            self.assertIsInstance(i, str)

    def test_reprContents_color(self):
        anchor = self.getAnchor_generic()
        anchor.color = (1, 0, 1, 1)
        value = anchor._reprContents()
        self.assertIsInstance(value, list)
        for i in value:
            self.assertIsInstance(i, str)

    def test_reprContents_noGlyph_color(self):
        anchor, _ = self.objectGenerator("anchor")
        anchor.color = (1, 0, 1, 1)
        value = anchor._reprContents()
        self.assertIsInstance(value, list)
        for i in value:
            self.assertIsInstance(i, str)

    # ----------
    # Attributes
    # ----------

    # Name

    def test_get(self):
        anchor = self.getAnchor_generic()
        self.assertEqual(anchor.name, "Anchor Attribute Test")

    def test_set_valid(self):
        anchor = self.getAnchor_generic()
        anchor.name = "foo"
        self.assertEqual(anchor.name, "foo")

    def test_set_none(self):
        anchor = self.getAnchor_generic()
        anchor.name = None
        self.assertIsNone(anchor.name)

    def test_set_invalid(self):
        anchor = self.getAnchor_generic()
        with self.assertRaises(TypeError):
            anchor.name = 123

    # Color

    def test_color_get_none(self):
        anchor = self.getAnchor_generic()
        self.assertIsNone(anchor.color)

    def test_color_set_valid_max(self):
        anchor = self.getAnchor_generic()
        anchor.color = (1, 1, 1, 1)
        self.assertEqual(anchor.color, (1, 1, 1, 1))

    def test_color_set_valid_min(self):
        anchor = self.getAnchor_generic()
        anchor.color = (0, 0, 0, 0)
        self.assertEqual(anchor.color, (0, 0, 0, 0))

    def test_color_set_valid_decimal(self):
        anchor = self.getAnchor_generic()
        anchor.color = (0.1, 0.2, 0.3, 0.4)
        self.assertEqual(anchor.color, (0.1, 0.2, 0.3, 0.4))

    def test_color_set_none(self):
        anchor = self.getAnchor_generic()
        anchor.color = None
        self.assertIsNone(anchor.color)

    def test_color_set_invalid_over_max(self):
        anchor = self.getAnchor_generic()
        with self.assertRaises(ValueError):
            anchor.color = (1.1, 0.2, 0.3, 0.4)

    def test_color_set_invalid_uner_min(self):
        anchor = self.getAnchor_generic()
        with self.assertRaises(ValueError):
            anchor.color = (-0.1, 0.2, 0.3, 0.4)

    def test_color_set_invalid_too_few(self):
        anchor = self.getAnchor_generic()
        with self.assertRaises(ValueError):
            anchor.color = (0.1, 0.2, 0.3)

    def test_color_set_invalid_string(self):
        anchor = self.getAnchor_generic()
        with self.assertRaises(TypeError):
            anchor.color = "0.1,0.2,0.3,0.4"

    def test_color_set_invalid_int(self):
        anchor = self.getAnchor_generic()
        with self.assertRaises(TypeError):
            anchor.color = 123

    # Identifier

    def test_identifier_get_none(self):
        anchor = self.getAnchor_generic()
        self.assertIsNone(anchor.identifier)

    def test_identifier_generated_type(self):
        anchor = self.getAnchor_generic()
        anchor.getIdentifier()
        self.assertIsInstance(anchor.identifier, str)

    def test_identifier_consistency(self):
        anchor = self.getAnchor_generic()
        anchor.getIdentifier()
        # get: twice to test consistency
        self.assertEqual(anchor.identifier, anchor.identifier)

    def test_identifier_cannot_set(self):
        # identifier is a read-only property
        anchor = self.getAnchor_generic()
        with self.assertRaises(FontPartsError):
            anchor.identifier = "ABC"

    def test_identifier_force_set(self):
        identifier = "ABC"
        anchor = self.getAnchor_generic()
        anchor._setIdentifier(identifier)
        self.assertEqual(anchor.identifier, identifier)

    # Index

    def getAnchor_index(self):
        glyph, _ = self.objectGenerator("glyph")
        glyph.appendAnchor("anchor 0", (0, 0))
        glyph.appendAnchor("anchor 1", (0, 0))
        glyph.appendAnchor("anchor 2", (0, 0))
        return glyph

    def test_get_index_noParent(self):
        anchor, _ = self.objectGenerator("anchor")
        self.assertIsNone(anchor.index)

    def test_get_index(self):
        glyph = self.getAnchor_index()
        for i, anchor in enumerate(glyph.anchors):
            self.assertEqual(anchor.index, i)

    def test_set_index_noParent(self):
        anchor, _ = self.objectGenerator("anchor")
        with self.assertRaises(FontPartsError):
            anchor.index = 1

    def test_set_index_positive(self):
        glyph = self.getAnchor_index()
        anchor = glyph.anchors[0]
        with self.assertRaises(FontPartsError):
            anchor.index = 2

    def test_set_index_negative(self):
        glyph = self.getAnchor_index()
        anchor = glyph.anchors[1]
        with self.assertRaises(FontPartsError):
            anchor.index = -1

    # x

    def test_x_get(self):
        anchor = self.getAnchor_generic()
        self.assertEqual(anchor.x, 1)

    def test_x_set_valid_positive(self):
        anchor = self.getAnchor_generic()
        anchor.x = 100
        self.assertEqual(anchor.x, 100)

    def test_x_set_valid_negative(self):
        anchor = self.getAnchor_generic()
        anchor.x = -100
        self.assertEqual(anchor.x, -100)

    def test_x_set_valid_zero(self):
        anchor = self.getAnchor_generic()
        anchor.x = 0
        self.assertEqual(anchor.x, 0)

    def test_x_set_valid_positive_decimal(self):
        anchor = self.getAnchor_generic()
        anchor.x = 1.1
        self.assertEqual(anchor.x, 1.1)

    def test_x_set_valid_negative_decimal(self):
        anchor = self.getAnchor_generic()
        anchor.x = -1.1
        self.assertEqual(anchor.x, -1.1)

    def test_x_set_invalid_none(self):
        anchor = self.getAnchor_generic()
        with self.assertRaises(TypeError):
            anchor.x = None

    def test_x_set_valid_string(self):
        anchor = self.getAnchor_generic()
        with self.assertRaises(TypeError):
            anchor.x = "ABC"

    # y

    def test_y_get(self):
        anchor = self.getAnchor_generic()
        self.assertEqual(anchor.y, 2)

    def test_y_set_valid_positive(self):
        anchor = self.getAnchor_generic()
        anchor.y = 100
        self.assertEqual(anchor.y, 100)

    def test_y_set_valid_negative(self):
        anchor = self.getAnchor_generic()
        anchor.y = -100
        self.assertEqual(anchor.y, -100)

    def test_y_set_valid_zero(self):
        anchor = self.getAnchor_generic()
        anchor.y = 0
        self.assertEqual(anchor.y, 0)

    def test_y_set_valid_positive_decimal(self):
        anchor = self.getAnchor_generic()
        anchor.y = 1.1
        self.assertEqual(anchor.y, 1.1)

    def test_y_set_valid_negative_decimal(self):
        anchor = self.getAnchor_generic()
        anchor.y = -1.1
        self.assertEqual(anchor.y, -1.1)

    def test_y_set_invalid_none(self):
        anchor = self.getAnchor_generic()
        with self.assertRaises(TypeError):
            anchor.y = None

    def test_y_set_valid_string(self):
        anchor = self.getAnchor_generic()
        with self.assertRaises(TypeError):
            anchor.y = "ABC"

    # -------
    # Methods
    # -------

    def getAnchor_copy(self):
        anchor = self.getAnchor_generic()
        anchor.color = (0.1, 0.2, 0.3, 0.4)
        return anchor

    # copy

    def test_copy_seperate_objects(self):
        anchor = self.getAnchor_copy()
        copied = anchor.copy()
        self.assertIsNot(anchor, copied)

    def test_copy_same_name(self):
        anchor = self.getAnchor_copy()
        copied = anchor.copy()
        self.assertEqual(anchor.name, copied.name)

    def test_copy_same_color(self):
        anchor = self.getAnchor_copy()
        copied = anchor.copy()
        self.assertEqual(anchor.color, copied.color)

    def test_copy_same_identifier(self):
        anchor = self.getAnchor_copy()
        copied = anchor.copy()
        self.assertEqual(anchor.identifier, copied.identifier)

    def test_copy_generated_identifier_different(self):
        anchor = self.getAnchor_copy()
        copied = anchor.copy()
        anchor.getIdentifier()
        copied.getIdentifier()
        self.assertNotEqual(anchor.identifier, copied.identifier)

    def test_copy_same_x(self):
        anchor = self.getAnchor_copy()
        copied = anchor.copy()
        self.assertEqual(anchor.x, copied.x)

    def test_copy_same_y(self):
        anchor = self.getAnchor_copy()
        copied = anchor.copy()
        self.assertEqual(anchor.y, copied.y)

    # transform

    def test_transformBy_valid_no_origin(self):
        anchor = self.getAnchor_generic()
        anchor.transformBy((2, 0, 0, 3, -3, 2))
        self.assertEqual(anchor.x, -1)
        self.assertEqual(anchor.y, 8)

    def test_transformBy_valid_origin(self):
        anchor = self.getAnchor_generic()
        anchor.transformBy((2, 0, 0, 2, 0, 0), origin=(1, 2))
        self.assertEqual(anchor.x, 1)
        self.assertEqual(anchor.y, 2)

    def test_transformBy_invalid_one_string_value(self):
        anchor = self.getAnchor_generic()
        with self.assertRaises(TypeError):
            anchor.transformBy((1, 0, 0, 1, 0, "0"))

    def test_transformBy_invalid_all_string_values(self):
        anchor = self.getAnchor_generic()
        with self.assertRaises(TypeError):
            anchor.transformBy("1, 0, 0, 1, 0, 0")

    def test_transformBy_invalid_int_value(self):
        anchor = self.getAnchor_generic()
        with self.assertRaises(TypeError):
            anchor.transformBy(123)

    # moveBy

    def test_moveBy_valid(self):
        anchor = self.getAnchor_generic()
        anchor.moveBy((-1, 2))
        self.assertEqual(anchor.x, 0)
        self.assertEqual(anchor.y, 4)

    def test_moveBy_invalid_one_string_value(self):
        anchor = self.getAnchor_generic()
        with self.assertRaises(TypeError):
            anchor.moveBy((-1, "2"))

    def test_moveBy_invalid_all_strings_value(self):
        anchor = self.getAnchor_generic()
        with self.assertRaises(TypeError):
            anchor.moveBy("-1, 2")

    def test_moveBy_invalid_int_value(self):
        anchor = self.getAnchor_generic()
        with self.assertRaises(TypeError):
            anchor.moveBy(1)

    # scaleBy

    def test_scaleBy_valid_one_value_no_origin(self):
        anchor = self.getAnchor_generic()
        anchor.scaleBy(-2)
        self.assertEqual(anchor.x, -2)
        self.assertEqual(anchor.y, -4)

    def test_scaleBy_valid_two_values_no_origin(self):
        anchor = self.getAnchor_generic()
        anchor.scaleBy((-2, 3))
        self.assertEqual(anchor.x, -2)
        self.assertEqual(anchor.y, 6)

    def test_scaleBy_valid_two_values_origin(self):
        anchor = self.getAnchor_generic()
        anchor.scaleBy((-2, 3), origin=(1, 2))
        self.assertEqual(anchor.x, 1)
        self.assertEqual(anchor.y, 2)

    def test_scaleBy_invalid_one_string_value(self):
        anchor = self.getAnchor_generic()
        with self.assertRaises(TypeError):
            anchor.scaleBy((-1, "2"))

    def test_scaleBy_invalid_two_string_values(self):
        anchor = self.getAnchor_generic()
        with self.assertRaises(TypeError):
            anchor.scaleBy("-1, 2")

    def test_scaleBy_invalid_tuple_too_many_values(self):
        anchor = self.getAnchor_generic()
        with self.assertRaises(ValueError):
            anchor.scaleBy((-1, 2, -3))

    # rotateBy

    def test_rotateBy_valid_no_origin(self):
        anchor = self.getAnchor_generic()
        anchor.rotateBy(45)
        self.assertAlmostEqual(anchor.x, -0.707, places=3)
        self.assertAlmostEqual(anchor.y, 2.121, places=3)

    def test_rotateBy_valid_origin(self):
        anchor = self.getAnchor_generic()
        anchor.rotateBy(45, origin=(1, 2))
        self.assertAlmostEqual(anchor.x, 1)
        self.assertAlmostEqual(anchor.y, 2)

    def test_rotateBy_invalid_string_value(self):
        anchor = self.getAnchor_generic()
        with self.assertRaises(TypeError):
            anchor.rotateBy("45")

    def test_rotateBy_invalid_too_large_value_positive(self):
        anchor = self.getAnchor_generic()
        with self.assertRaises(ValueError):
            anchor.rotateBy(361)

    def test_rotateBy_invalid_too_large_value_negative(self):
        anchor = self.getAnchor_generic()
        with self.assertRaises(ValueError):
            anchor.rotateBy(-361)

    # skewBy

    def test_skewBy_valid_no_origin_one_value(self):
        anchor = self.getAnchor_generic()
        anchor.skewBy(100)
        self.assertAlmostEqual(anchor.x, -10.343, places=3)
        self.assertEqual(anchor.y, 2.0)

    def test_skewBy_valid_no_origin_two_values(self):
        anchor = self.getAnchor_generic()
        anchor.skewBy((100, 200))
        self.assertAlmostEqual(anchor.x, -10.343, places=3)
        self.assertAlmostEqual(anchor.y, 2.364, places=3)

    def test_skewBy_valid_origin_one_value(self):
        anchor = self.getAnchor_generic()
        anchor.skewBy(100, origin=(1, 2))
        self.assertEqual(anchor.x, 1)
        self.assertEqual(anchor.y, 2)

    def test_skewBy_valid_origin_two_values(self):
        anchor = self.getAnchor_generic()
        anchor.skewBy((100, 200), origin=(1, 2))
        self.assertEqual(anchor.x, 1)
        self.assertEqual(anchor.y, 2)

    # round

    def getAnchor_round(self):
        anchor = self.getAnchor_generic()
        anchor.x = 1.1
        anchor.y = 2.5
        return anchor

    def test_round_close_to(self):
        anchor = self.getAnchor_round()
        anchor.round()
        self.assertEqual(anchor.x, 1)

    def test_round_at_half(self):
        anchor = self.getAnchor_round()
        anchor.round()
        self.assertEqual(anchor.y, 3)

    # ----
    # Hash
    # ----

    def test_hash_object_self(self):
        anchor_one = self.getAnchor_generic()
        self.assertEqual(hash(anchor_one), hash(anchor_one))

    def test_hash_object_other(self):
        anchor_one = self.getAnchor_generic()
        anchor_two = self.getAnchor_generic()
        self.assertNotEqual(hash(anchor_one), hash(anchor_two))

    def test_hash_object_self_variable_assignment(self):
        anchor_one = self.getAnchor_generic()
        a = anchor_one
        self.assertEqual(hash(anchor_one), hash(a))

    def test_hash_object_other_variable_assignment(self):
        anchor_one = self.getAnchor_generic()
        anchor_two = self.getAnchor_generic()
        a = anchor_one
        self.assertNotEqual(hash(anchor_two), hash(a))

    def test_is_hashable(self):
        anchor_one = self.getAnchor_generic()
        self.assertTrue(isinstance(anchor_one, collections.abc.Hashable))

    # -------
    # Parents
    # -------

    def test_get_parent_font(self):
        font, _ = self.objectGenerator("font")
        layer = font.newLayer("L")
        glyph = layer.newGlyph("X")
        anchor = glyph.appendAnchor("anchor 0", (0, 0))
        self.assertIsNotNone(anchor.font)
        self.assertEqual(anchor.font, font)

    def test_get_parent_noFont(self):
        layer, _ = self.objectGenerator("layer")
        glyph = layer.newGlyph("X")
        anchor = glyph.appendAnchor("anchor 0", (0, 0))
        self.assertIsNone(anchor.font)

    def test_get_parent_layer(self):
        layer, _ = self.objectGenerator("layer")
        glyph = layer.newGlyph("X")
        anchor = glyph.appendAnchor("anchor 0", (0, 0))
        self.assertIsNotNone(anchor.layer)
        self.assertEqual(anchor.layer, layer)

    def test_get_parent_noLayer(self):
        glyph, _ = self.objectGenerator("glyph")
        anchor = glyph.appendAnchor("anchor 0", (0, 0))
        self.assertIsNone(anchor.font)
        self.assertIsNone(anchor.layer)

    def test_get_parent_glyph(self):
        glyph, _ = self.objectGenerator("glyph")
        anchor = glyph.appendAnchor("anchor 0", (0, 0))
        self.assertIsNotNone(anchor.glyph)
        self.assertEqual(anchor.glyph, glyph)

    def test_get_parent_noGlyph(self):
        anchor, _ = self.objectGenerator("anchor")
        self.assertIsNone(anchor.font)
        self.assertIsNone(anchor.layer)
        self.assertIsNone(anchor.glyph)

    def test_set_parent_glyph(self):
        glyph, _ = self.objectGenerator("glyph")
        anchor = self.getAnchor_generic()
        anchor.glyph = glyph
        self.assertIsNotNone(anchor.glyph)
        self.assertEqual(anchor.glyph, glyph)

    def test_set_parent_glyph_none(self):
        anchor, _ = self.objectGenerator("anchor")
        anchor.glyph = None
        self.assertIsNone(anchor.glyph)

    def test_set_parent_glyph_exists(self):
        glyph, _ = self.objectGenerator("glyph")
        otherGlyph, _ = self.objectGenerator("glyph")
        anchor = glyph.appendAnchor("anchor 0", (0, 0))
        with self.assertRaises(AssertionError):
            anchor.glyph = otherGlyph

    # --------
    # Equality
    # --------

    def test_object_equal_self(self):
        anchor_one = self.getAnchor_generic()
        self.assertEqual(anchor_one, anchor_one)

    def test_object_not_equal_other(self):
        anchor_one = self.getAnchor_generic()
        anchor_two = self.getAnchor_generic()
        self.assertNotEqual(anchor_one, anchor_two)

    def test_object_equal_variable_assignment_self(self):
        anchor_one = self.getAnchor_generic()
        a = anchor_one
        a.moveBy((-1, 2))
        self.assertEqual(anchor_one, a)

    def test_object_not_equal_variable_assignment_other(self):
        anchor_one = self.getAnchor_generic()
        anchor_two = self.getAnchor_generic()
        a = anchor_one
        self.assertNotEqual(anchor_two, a)

    # ---------
    # Selection
    # ---------

    def test_selected_true(self):
        anchor = self.getAnchor_generic()
        try:
            anchor.selected = False
        except NotImplementedError:
            return
        anchor.selected = True
        self.assertEqual(anchor.selected, True)

    def test_selected_false(self):
        anchor = self.getAnchor_generic()
        try:
            anchor.selected = False
        except NotImplementedError:
            return
        self.assertEqual(anchor.selected, False)



from __future__ import annotations
from collections import namedtuple
from typing import overload, Any

from fontParts.base.normalizers import normalizeColor
from fontParts.base.annotations import IntFloatType, RGBALike

_BaseColor = namedtuple("_BaseColor", ["r", "g", "b", "a"])


class Color(_BaseColor):
    """Represent a color object following the :ref:`type-color`.

    This class accepts either four RGBA component values or a :class:`list` or
    :class:`tuple` of these values.

    :param r: The color's red component as :class:`int` or :class:`float`.
    :param g: The color's green component as :class:`int` or :class:`float`.
    :param b: The color's blue component as :class:`int` or :class:`float`.
    :param a: The color's alpha component as :class:`int` or :class:`float`.

    """

    __slots__ = ()

    @overload
    def __new__(
        cls, r: IntFloatType, g: IntFloatType, b: IntFloatType, a: IntFloatType
    ) -> Color: ...

    @overload
    def __new__(cls, color: RGBALike) -> Color: ...

    def __new__(cls, *args: Any, **kwargs: Any) -> Color:
        if kwargs:
            value = (
                kwargs.get("r", 0),
                kwargs.get("g", 0),
                kwargs.get("b", 0),
                kwargs.get("a", 1),
            )

        else:
            value = args[0] if len(args) == 1 else args
        normalizedValue = normalizeColor(value)  # type: ignore[arg-type]
        return super().__new__(cls, *normalizedValue)

    # Remove auto-generated _BaseCpøpr.__new__ docstring
    __new__.__doc__ = ""



import unittest
import collections


class TestGroups(unittest.TestCase):
    def getGroups_generic(self):
        groups, _ = self.objectGenerator("groups")
        groups.update(
            {
                "group 1": ("A", "B", "C"),
                "group 2": ("x", "y", "z"),
                "group 3": (),
                "group 4": ("A",),
            }
        )
        return groups

    # ----
    # repr
    # ----

    def test_reprContents(self):
        font, _ = self.objectGenerator("font")
        groups = font.groups
        value = groups._reprContents()
        self.assertIsInstance(value, list)
        found = False
        for i in value:
            self.assertIsInstance(i, str)
            if "for font" in value:
                found = True
        self.assertTrue(found)

    def test_reprContents_noFont(self):
        groups, _ = self.objectGenerator("groups")
        value = groups._reprContents()
        self.assertIsInstance(value, list)
        self.assertEqual(value, [])

    # -------
    # Parents
    # -------

    def test_get_parent_font(self):
        font, _ = self.objectGenerator("font")
        groups = font.groups
        self.assertIsNotNone(groups.font)
        self.assertEqual(groups.font, font)

    def test_get_parent_font_none(self):
        groups, _ = self.objectGenerator("groups")
        self.assertIsNone(groups.font)

    def test_set_parent_font(self):
        font, _ = self.objectGenerator("font")
        groups, _ = self.objectGenerator("groups")
        groups.font = font
        self.assertIsNotNone(groups.font)
        self.assertEqual(groups.font, font)

    def test_set_parent_font_none(self):
        groups, _ = self.objectGenerator("groups")
        groups.font = None
        self.assertIsNone(groups.font)

    def test_set_parent_differentFont(self):
        font, _ = self.objectGenerator("font")
        fontB, _ = self.objectGenerator("font")
        groups, _ = self.objectGenerator("groups")
        groups.font = font
        self.assertIsNotNone(groups.font)

        with self.assertRaises(AssertionError):
            groups.font = fontB

    # ---
    # len
    # ---

    def test_len_initial(self):
        groups = self.getGroups_generic()
        self.assertEqual(len(groups), 4)

    def test_len_clear(self):
        groups = self.getGroups_generic()
        groups.clear()
        self.assertEqual(len(groups), 0)

    def test_len_add(self):
        groups = self.getGroups_generic()
        groups["group 5"] = ["D", "E", "F"]
        self.assertEqual(len(groups), 5)

    def test_len_subtract(self):
        groups = self.getGroups_generic()
        groups.pop("group 4")
        self.assertEqual(len(groups), 3)

    # ---
    # Get
    # ---

    def test_get_fallback_default(self):
        groups = self.getGroups_generic()
        self.assertEqual(groups.get("test"), None)

    # -------
    # Queries
    # -------

    def test_find_found(self):
        groups = self.getGroups_generic()
        found = groups.findGlyph("A")
        found.sort()
        self.assertEqual(found, ["group 1", "group 4"])

    def test_find_not_found(self):
        groups = self.getGroups_generic()
        self.assertEqual(groups.findGlyph("five"), [])

    def test_find_invalid_key(self):
        groups = self.getGroups_generic()
        with self.assertRaises(TypeError):
            groups.findGlyph(5)

    def test_contains_found(self):
        groups = self.getGroups_generic()
        self.assertTrue("group 4" in groups)

    def test_contains_not_found(self):
        groups = self.getGroups_generic()
        self.assertFalse("group five" in groups)

    def test_get_found(self):
        groups = self.getGroups_generic()
        self.assertEqual(groups["group 1"], ("A", "B", "C"))

    def test_get_not_found(self):
        groups = self.getGroups_generic()
        with self.assertRaises(KeyError):
            groups["group two"]

    # --------------
    # Kerning Groups
    # --------------

    def getGroups_kerning(self):
        groups = self.getGroups_generic()
        kerningGroups = {
            "public.kern1.A": ("A", "Aacute"),
            "public.kern1.O": ("O", "D"),
            "public.kern2.A": ("A", "Aacute"),
            "public.kern2.O": ("O", "C"),
        }
        groups.update(kerningGroups)
        return groups

    def test_side1KerningGroups(self):
        groups = self.getGroups_kerning()
        expected = {"public.kern1.A": ("A", "Aacute"), "public.kern1.O": ("O", "D")}
        self.assertEqual(groups.side1KerningGroups, expected)
        # self.assertEqual(super(groups, self)._get_side1KerningGroups(), expected)

    def test_get_side1KerningGroups(self):
        groups = self.getGroups_kerning()
        expected = {"public.kern1.A": ("A", "Aacute"), "public.kern1.O": ("O", "D")}
        self.assertEqual(groups._get_side1KerningGroups(), expected)

    def test_side2KerningGroups(self):
        groups = self.getGroups_kerning()
        expected = {"public.kern2.A": ("A", "Aacute"), "public.kern2.O": ("O", "C")}
        self.assertEqual(groups.side2KerningGroups, expected)

    def test_get_side2KerningGroups(self):
        groups = self.getGroups_kerning()
        expected = {"public.kern1.A": ("A", "Aacute"), "public.kern1.O": ("O", "D")}
        self.assertEqual(groups._get_side1KerningGroups(), expected)

    # ----
    # Hash
    # ----

    def test_hash(self):
        groups = self.getGroups_generic()
        self.assertEqual(isinstance(groups, collections.abc.Hashable), True)

    # --------
    # Equality
    # --------

    def test_object_equal_self(self):
        groups_one = self.getGroups_generic()
        self.assertEqual(groups_one, groups_one)

    def test_object_not_equal_other(self):
        groups_one = self.getGroups_generic()
        groups_two = self.getGroups_generic()
        self.assertNotEqual(groups_one, groups_two)

    def test_object_equal_self_variable_assignment(self):
        groups_one = self.getGroups_generic()
        a = groups_one
        self.assertEqual(groups_one, a)

    def test_object_not_equal_other_variable_assignment(self):
        groups_one = self.getGroups_generic()
        groups_two = self.getGroups_generic()
        a = groups_one
        self.assertNotEqual(groups_two, a)

    # ---------------------
    # RoboFab Compatibility
    # ---------------------

    def test_remove(self):
        groups = self.getGroups_generic()
        groups.remove("group 2")
        expected = {"group 1": ("A", "B", "C"), "group 3": (), "group 4": ("A",)}
        self.assertEqual(groups.asDict(), expected)

    def test_remove_twice(self):
        groups = self.getGroups_generic()
        groups.remove("group 1")
        with self.assertRaises(KeyError):
            groups.remove("group 1")

    def test_remove_nonexistant_group(self):
        groups = self.getGroups_generic()
        with self.assertRaises(KeyError):
            groups.remove("group 7")

    def test_asDict(self):
        groups = self.getGroups_generic()
        expected = {
            "group 1": ("A", "B", "C"),
            "group 2": ("x", "y", "z"),
            "group 3": (),
            "group 4": ("A",),
        }
        self.assertEqual(groups.asDict(), expected)

    # -------------------
    # Inherited Functions
    # -------------------

    def test_iter(self):
        groups = self.getGroups_generic()
        expected = ["group 1", "group 2", "group 3", "group 4"]

        listOfGroups = []
        for groupName in groups:
            listOfGroups.append(groupName)

        self.assertEqual(listOfGroups.sort(), expected.sort())

    def test_iter_remove(self):
        groups = self.getGroups_generic()

        for groupName in groups:
            groups.remove(groupName)
            self.assertNotIn(groupName, groups)

        self.assertFalse(groups)

    def test_values(self):
        groups = self.getGroups_generic()
        expected = [("A", "B", "C"), ("x", "y", "z"), (), ("A",)]
        self.assertEqual(sorted(groups.values()), sorted(expected))
